#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# August 2025
#
import os
import contextlib
import numpy as np
import tqdm

import veux
import xara
from xcsi.csi import collect_outlines, load as load_csi
from xcsi.job import Job
from xcsi.metrics import PeakDrift


from dataclasses import dataclass

@dataclass
class Datum:
    angle_x: float = 0.0
    angle_y: float = 0.0


    def to_cardinal(self):
        """
        Create rotation matrix from datum basis to cardinal basis.
        The cardinal basis is defined as:
        cx = North, 
        cy = East, 
        cz = Up

        The datum basis is defined as:
        ex = Exp(orient_x cz) cx
        ey = Exp(orient_y cz) cy
        ez = Exp(orient_z cz) cz
        where orient_x, orient_y, and orient_z are the angles in radians and
        Exp is the exponential map which encodes the Rodrigues' rotation formula.

        We want to return the rotation Rec  such that:
            Rec @ [ex, ey, ez] = [cx, cy, cz]
        """
        import numpy as np
        from shps.rotor import exp

        orient_x = self.angle_x
        orient_y = self.angle_y

        ex, ey, ez = np.eye(3)
        cz = ez
        cx = exp(-orient_x*cz) @ ex
        cy = exp(-orient_y*cz) @ ey

        rotation = np.column_stack((cx, cy, cz))

        return rotation
    
    def to_other(self, other: "Datum"):
        """
        Create rotation matrix from datum basis to other datum basis.
        """
        R_cs = self.to_cardinal()
        R_cm = other.to_cardinal()
        return R_cm.T @ R_cs



def _analyze_and_render(model, artist, nt, dt):
    import veux, veux.motion

    motion = veux.motion.Motion(artist)
    for i in tqdm.tqdm(range(nt)):
        if model.analyze(1, dt) != 0:
            return -1
        motion.advance(i*dt)
        motion.draw_sections(position=lambda x: [1000*u for u in model.nodeDisp(x)],
                             rotation=model.nodeRotation)
        
    motion.add_to(artist.canvas)
    return 0


@contextlib.contextmanager
def new_cd(x):
    d = os.getcwd()
    # This could raise an exception, but it's probably
    # best to let it propagate and let the caller
    # deal with it, since they requested x
    os.chdir(x)

    try:
        yield

    finally:
        # This could also raise an exception, but you *really*
        # aren't equipped to figure out what went wrong if the
        # old working directory can't be restored.
        os.chdir(d)


def _create_excitation(asm, inputs, dt):
    import numpy as np
    model = asm.model
    units = asm.units

    if units is None:
        raise ValueError("Units not defined in assembly")

    rotation = np.eye(3)
    i = 1
    for sensor in inputs.values():
        if sensor["role"] == "input":
            for dof in range(3):
                series = sum(ai*dx*units.cm
                            for ai, dx in zip(np.array(sensor["series"]), rotation[dof]))

                model.timeSeries("Path", i, dt=dt, values=series.tolist())
                model.pattern("UniformExcitation", i, dof+1, accel=i)
                i += 1



def _create_inputs(event, config):

    nt = 0
    dt = 0.02

    inputs = {}
    R_ms = Datum(**config["sensor_group"]["datum"]).to_other(Datum(**config["predictor"]["datum"]))

    for sensor_assign in config["sensor_assign"]:
        sensor = next((s for s in config["sensor_group"]["sensors"]
                       if s["id"] == sensor_assign["id"]), None)
        if sensor is None:
            raise ValueError(f"Sensor {sensor_assign['id']} not found in sensor group")

        sensor_series = event.match("l", station_channel=str(sensor["id"])).accel.data
        # Series in the Sensor coordinate system
        series_s = [
            dx*sensor_series for dx in sensor["orient"]
        ]
        # Series in the Model coordinate system
        series_m = [
            sum(mij*si for mij in mi).tolist() for mi, si in zip(R_ms.T, series_s)
        ]

        inputs[sensor["id"]] = {
            "role":  "input",
            "series": series_m
        }
        nt = len(series_m[0])


    return inputs, nt, dt



def runPrediction(csi, event, config):

    # 1) Event
    nt = 500
    dt = 0.02
    inputs, nt, dt = _create_inputs(event, config)


    # Create model

    with new_cd(config["directory"]), open("model.tcl", "w") as echo_file:

        model = xara.Model(ndm=3, ndf=6, echo_file=echo_file)
        asm = Job(csi).assemble(model=model)

        sections = collect_outlines(csi, model.frame_tags)

        #
        # DAMPING
        #

        #
        # DYNAMIC RECORDERS
        #

        ## COLUMN SECTION DEFORMATIONS AT TOP AND BOTTOM FOR STRAIN-BASED DAMAGE STATES
        if False:
            column_strains = tuple(k["key"] for k in self.runs[run_id]["columns"] if k["strain"])
            if len(column_strains) > 0:
                model.recorder("Element",  "section", 1, "deformation", xml="eleDef1.txt", ele=column_strains) # section 1 deformation]
                model.recorder("Element",  "section", 4, "deformation", xml="eleDef4.txt", ele=column_strains) # section 4 deformation]

        # RESPONSE HISTORY RECORDERS
        metrics = [
            PeakDrift((31, 81))
        ]

        for metric in metrics:
            metric.record(asm)


        #
        # Run dynamic analysis
        #
        _create_excitation(asm, inputs, dt)
        
        model.eval(f"print -json -file model.json")

        model.eval(f"""
        wipeAnalysis
        set NewmarkGamma    0.50;
        set NewmarkBeta     0.25;
        constraints Transformation;
        numberer    RCM;
        test        EnergyIncr 1.0e-8 50 0;
        system      Umfpack;
        integrator  Newmark $NewmarkGamma $NewmarkBeta;
        algorithm   Newton;
        analysis    Transient;
        """)

        artist = veux.create_artist(model, vertical=3, model_config={
                "frame_outlines": sections
        })
        _analyze_and_render(model, artist, nt, dt)

        artist.save("motion.glb")

        model.wipe()


if __name__ == "__main__":
    import quakeio 
    import sys 

    config = {
        "directory": "./0/",

        "sensor_group": {
            "datum":  {"angle_x": 0, "angle_y": 0},
            "sensors": [
                {"id": 1, "orient": [1.0, 0.0, 0.0]},
                {"id": 3, "orient": [0.0, 1.0, 0.0]}
            ]
        },

        # Predictor
        "sensor_assign": [
            {"id": 1, "role": "input"},
            {"id": 3, "role": "input"},
        ],

        "metrics": [
            {"type": "PeakDrift", "joints": [31, 81]},
            {"type": "VeuxMotion"}
        ],

        "predictor": {
            "datum": {"angle_x": 0, "angle_y": 0},
        }
    }


    with open(sys.argv[1], "r") as f:
        csi = load_csi(f)

    event = quakeio.read(sys.argv[2])

    runPrediction(csi, event, config)



import re
from collections import defaultdict

RE = {
    "joint_key": re.compile("Joint[0-9]")
}

TYPES = {
    "Shell": {
        "Elastic": "ShellMITC4",
    },
    "Frame": {
        "Elastic": "PrismFrame"
    }
}

class Converter:
    def __init__(self):
        self._log = []
        self._csi_names = {
            # Abaqus
            "Material": {
                "material": {},
                "uniaxial": {}
            },
            # CSI
            "Joint": {
                "node": {
                    # 1:    1,
                    # "P2": 2,
                    # 3:    3
                },
            },
            "Shell": {"element": {}},
            "Link":  {"element": {}},
            "Frame": {"element": {}},
            "AnalSect": {
                "section": {},
                "integration": {}
            },
            "ShellSection": {
                "section": {}
            }
        }
        self._ops_count = {
                "node":       0,
                "transform":  0,
                "element":    0,
                "nDMaterial": 0,
                "section":    0,
                "material":   0,
                "integration": 0
        }

        self._library = {
            "frame_sections": {},
            "shell_sections": {},
            "link_materials": defaultdict(dict),
        }

    def identify(self, csi_type, ops_type, csi_name)->int:
        if ops_type not in self._csi_names[csi_type]:
            self._csi_names[csi_type][ops_type] = {}
            return None

        if csi_name in self._csi_names[csi_type][ops_type]:
            return self._csi_names[csi_type][ops_type][csi_name]

        return None

    def define(self, csi_type, ops_type, csi_name=None, item=None)->int:
        if ops_type not in self._csi_names[csi_type]:
            self._csi_names[csi_type][ops_type] = {}

        if csi_name is None:
            self._ops_count[ops_type] += 1
            return self._ops_count[ops_type]

        if csi_name not in self._csi_names[csi_type][ops_type]:
            self._ops_count[ops_type] += 1
            if item is None:
                item = self._ops_count[ops_type]
            self._csi_names[csi_type][ops_type][csi_name] = item

        return self._csi_names[csi_type][ops_type][csi_name]


    def log(self, message):
        self._log.append(message)




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
from .utility import UnimplementedInstance, find_row
from ..names import RE, TYPES


def add_shells(csi, model, conv):
    for shell in csi.get("CONNECTIVITY - AREA", []):
        if "AREA ADDED MASS ASSIGNMENTS" in csi:
            row = find_row(csi["AREA ADDED MASS ASSIGNMENTS"],
                           Area=shell["Area"])
            if row:
                mass = row["MassPerArea"]
            else:
                mass = 0.0
        else:
            mass = 0.0

        # Find section
        assign  = find_row(csi["AREA SECTION ASSIGNMENTS"],
                           Area=shell["Area"])

        section = conv.identify("ShellSection", "section", assign["Section"])

        nodes = tuple(
            conv.identify("Joint", "node", v)
            for k,v in shell.items() if RE["joint_key"].match(k)
        )

        if len(nodes) == 4:
            type = TYPES["Shell"]["Elastic"]

        elif len(nodes) == 3:
            type = "ShellNLDKGT"

        model.element(type,
                      conv.define("Shell", "element", shell["Area"]),
                      nodes, section
        )



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import re
from collections import defaultdict

RE = {
    "joint_key": re.compile("Joint[0-9]")
}

TYPES = {
    "Shell": {
        "Elastic": "ShellMITC4",
    },
    "Frame": {
        "Elastic": "PrismFrame"
    }
}

class Names:
    def __init__(self):
        self._log = []
        self._csi_names = {
            # CSI
            "Joint": {
                "node": {},
            },
            "Shell": {"element": {}},
            "Link":  {"element": {}},
            "Frame": {"element": {}},
            "AnalSect": {
                "section": {},
                "integration": {}
            },
            "ShellSection": {
                "section": {}
            },

            "Material": {
                "material": {},
                "uniaxial": {}
            },

            "LoadPattern": {
                "pattern": {}
            },
        }


        self._ops_count = {
                "node":        0,
                "transform":   0,
                "element":     0,
                "nDMaterial":  0,
                "section":     0,
                "material":    0,
                "integration": 0,
                "pattern":     0,
        }

        self._library = {
            "frame_sections": {},
            "shell_sections": {},
            "link_materials": defaultdict(dict),
        }

    def identify(self, csi_type, ops_type, csi_name)->int:
        if ops_type not in self._csi_names[csi_type]:
            self._csi_names[csi_type][ops_type] = {}
            return None

        if csi_name in self._csi_names[csi_type][ops_type]:
            return self._csi_names[csi_type][ops_type][csi_name]

        return None

    def define(self, csi_type, ops_type, csi_name=None, item=None)->int:
        if ops_type not in self._csi_names[csi_type]:
            self._csi_names[csi_type][ops_type] = {}

        if csi_name is None:
            self._ops_count[ops_type] += 1
            return self._ops_count[ops_type]

        if csi_name not in self._csi_names[csi_type][ops_type]:
            self._ops_count[ops_type] += 1
            if item is None:
                item = self._ops_count[ops_type]
            self._csi_names[csi_type][ops_type][csi_name] = item

        return self._csi_names[csi_type][ops_type][csi_name]


    def log(self, message):
        self._log.append(message)




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import sys
import numpy as np
import warnings

from ..utility import UnimplementedInstance, find_row


def _is_truss(frame, csi):
    if "FRAME RELEASE ASSIGNMENTS 1 - GENERAL" in csi:
        release = find_row(csi["FRAME RELEASE ASSIGNMENTS 1 - GENERAL"],
                        Frame=frame["Frame"])
    else:
        return False

    return release and all(release[i] for i in ("TI", "M2I", "M3I", "M2J", "M3J"))


def _orient(xi, xj, angle):
    """
    Calculate the coordinate transformation vector.
    xi is the location of node I, xj node J,
    and `angle` is the rotation about the local axis in degrees

    By default local axis 2 is always in the 1-Z plane, except if the object
    is vertical and then it is parallel to the global X axis.
    The definition of the local axes follows the right-hand rule.
    """

    # The local 1 axis points from node I to node J
    dx, dy, dz = e1 = xj - xi
    # Global z
    E3 = np.array([0, 0, 1])

    # In Sap2000, if the element is vertical, the local y-axis is the same as the
    # global x-axis, and the local z-axis can be obtained by cross-multiplying
    # the local x-axis with the local y-axis.
    if dx == 0 and dy == 0:
        e2 = np.array([1, 0, 0])

    # Otherwise, the plane composed of the local x-axis and the local
    # y-axis is a vertical plane. In this
    # case, the local z-axis can be obtained by the cross product of the local
    # x-axis and the global z-axis.
    else:
        e2 = np.cross(E3, e1)

    e3 = np.cross(e1, e2)

    # Rotate the local axis using the Rodrigue rotation formula
    # convert from degrees to radians
    angle = angle / 180 * np.pi
    e3r = e3 * np.cos(angle) + np.cross(e1, e3) * np.sin(angle)
    # Finally, the normalized local z-axis is returned
    return e3r / np.linalg.norm(e3r)


def add_frames(csi, model, library, config, conv):
    ndm = config.get("ndm", 3)
    log = []

    # itag = 1
    transform = 1

    tags = {}

    for frame in csi.get("CONNECTIVITY - FRAME",[]):
        if _is_truss(frame, csi):
            conv.log(UnimplementedInstance("Truss", frame))
            continue

        if "IsCurved" in frame and frame["IsCurved"]:
            conv.log(UnimplementedInstance("Frame.Curve", frame))

        nodes = (
            conv.identify("Joint", "node", frame["JointI"]),
            conv.identify("Joint", "node", frame["JointJ"])
        )

        if "FRAME ADDED MASS ASSIGNMENTS" in csi:
            row = find_row(csi["FRAME ADDED MASS ASSIGNMENTS"],
                            Frame=frame["Frame"])
            additional_mass = row["MassPerLen"] if row else 0.0
        else:
            additional_mass = 0.0

        #
        # Geometric transformation
        #
        if "FRAME LOCAL AXES ASSIGNMENTS 1 - TYPICAL" in csi:
            row = find_row(csi["FRAME LOCAL AXES ASSIGNMENTS 1 - TYPICAL"],
                            Frame=frame["Frame"])
            angle = row["Angle"] if row else 0.0
        else:
            angle = 0

        xi = np.array(model.nodeCoord(nodes[0]))
        xj = np.array(model.nodeCoord(nodes[1]))
        if np.linalg.norm(xj - xi) < 1e-10:
            log.append(UnimplementedInstance("Frame.ZeroLength", frame))
            print(f"ZERO LENGTH FRAME: {frame['Frame']}", file=sys.stderr)
            continue

        if ndm == 3:
            vecxz = _orient(xi, xj, angle)
            model.geomTransf("Linear", transform, *vecxz)
        else:
            model.geomTransf("Linear", transform)

        transform += 1

        #
        # Section
        #
        assign = find_row(csi["FRAME SECTION ASSIGNMENTS"], Frame=frame["Frame"])
        
        sect_info = find_row(csi["FRAME SECTION PROPERTIES 01 - GENERAL"],
                             SectionName=assign["AnalSect"])
        if not sect_info:
            conv.log(UnimplementedInstance("FrameSection.Unknown", assign))
            continue

        # ---------------------------------------------------------------------
        # Handle prismatic vs nonprismatic to get mass/length
        # ---------------------------------------------------------------------
        is_nonprismatic = (sect_info["Shape"] == "Nonprismatic")

        if not is_nonprismatic:
            #   
            # Prismatic section
            #
            A = sect_info["Area"]  # cross‐sectional area
            mat_info = find_row(csi["MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES"],
                                Material=sect_info["Material"])
            rho = mat_info["UnitMass"]  # material density (mass / volume)

            # self‐weight mass per length = area × density
            self_weight_mpl = A * rho

        else:
            #
            # Nonprismatic section
            #
            # look for total mass and total length in
            # 

            # total mass for the entire nonprismatic section
            total_mass = sect_info.get("TotalMass", 0)
            if "NPSectLen" in assign:
                total_length = assign["NPSectLen"]
            else:
                # handle the case where NPSectLen doesn't exist
                total_length = np.linalg.norm(xj - xi)

            if total_length < 1e-10:
                conv.log(UnimplementedInstance("FrameSection.NonprismaticZeroLength", assign))
                continue

            # self‐weight mass per length = total mass / total length
            self_weight_mpl = total_mass / total_length

        # Combine any assigned mass per length with self‐weight mass per length
        total_mass = self_weight_mpl + additional_mass

        # section = library["frame_sections"][assign["AnalSect"]] # conv.identify("AnalSect", "section", assign["AnalSect"]) #

        # TODO: probably dont use "SectionType" from this table; it looks like its 
        # superceded by "Shape" in Section Properties 01 table.
        if ("SectionType" not in assign) or (assign["SectionType"] != "Nonprismatic") or \
           assign["NPSectType"] == "Advanced":

            section = conv.identify("AnalSect", "section", assign["AnalSect"])
            if section is None:
                warnings.warn(f"No section found for {assign['AnalSect']}")
                continue

            e = model.element("PrismFrame",
                          conv.define("Frame", "element", frame["Frame"]),
                          nodes,
                          section=section,
                          transform=transform-1,
                          mass=total_mass
            )
            tags[frame["Frame"]] = e


        elif assign["NPSectType"] == "Default" and (integr := conv.identify("AnalSect", "integration", assign["AnalSect"])):
            # Non-prismatic sections
            e = model.element("ForceFrame",
                              conv.define("Frame", "element", frame["Frame"]),
                              nodes,
                              transform-1,
                              integr,
                              mass=total_mass
            )
            tags[frame["Frame"]] = e

        else:
            conv.log(UnimplementedInstance("FrameSection.NPSectType", assign["NPSectType"]))
            continue

    library["frame_tags"] = tags


    return log





from dataclasses import dataclass

class _Pattern:
    @dataclass
    class JointLoad:
        joint: int
        force: list

    @dataclass
    class FrameLoad:
        pass

    def __init__(self):
        self.joint_forces = []
        self.frame_forces = []
        self.shell_forces = []

def create_loads(csi, types=None, verbose=False):
    patterns = {}
    for pattern in csi.get("LOAD PATTERN DEFINITIONS", []):
        patterns[pattern["LoadPat"]] = _Pattern()

    "FRAME LOADS - DISTRIBUTED"  # 1-001
    "FRAME LOADS - POINT"        # 1-001
    "JOINT LOADS - FORCE"







class Log:
    def append(self, table, field, object):
        pass

class UnimplementedInstance:
    def __init__(self, name, object=None, table=None):
        self.name = name 
        self.object = object
        self.table = table

    def __repr__(self):
        return f"{self.name}: {self.object}"

def print_log(log):
    import sys

    types = {i.name for i in log}

    print("Unimplemented features", file=sys.stderr)
    for item in types:
        print(f"\t{item}: {sum(1 for i in log if i.name == item)}", file=sys.stderr)

def find_row(table, **kwds) -> dict:

    for row in table:
        match = True
        for k, v in kwds.items():
            if k not in row or row[k] != v:
                match = False
                break

        if match:
            return row


def find_rows(table, **kwds) -> list:
    rows = []
    for row in table:
        match = True
        for k, v in kwds.items():
            if k not in row or row[k] != v:
                match = False
                break

        if match:
            rows.append(row)

    return rows




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import numpy as np
import warnings
from .utility import UnimplementedInstance, find_row, find_rows
from .handler import Handler


class LinkHandler(Handler):
    pass

def _orient(xi, xj, deg):
    # Calculate the direction vector of the link element
    # Where a is the node number of node i, b is the node number of node j, and degree is the user-specified local axis
    # ------------------------------------------------------------------------------
    d_x, d_y, d_z = e_x = xj - xi

    # Local 1-axis points from node I to node J
    l_x = np.array([d_x, d_y, d_z])
    # Global z-axis
    g_z = np.array([0, 0, 1])

    # In SAP2000, if the link is vertical, the local y-axis is the same as the
    # global x-axis, and the local z-axis can be obtained by crossing the local
    # x-axis with the local y-axis
    if d_x == 0 and d_y == 0:
        l_y = np.array([1, 0, 0])
        l_z = np.cross(l_x, l_y)

    # In other cases, the plane formed by the local x-axis and the local y-axis
    # is a vertical plane (i.e., the normal vector is horizontal), and the
    # local z-axis can be obtained by crossing the local x-axis with the global
    # z-axis
    else:
        l_z = np.cross(l_x, g_z)

    # The local axis may also be rotated using the Rodrigues' rotation formula
    angle = deg / 180 * np.pi
    l_z_rot = l_z * np.cos(angle) + np.cross(l_x, l_z) * np.sin(angle)
    # The rotated local y-axis can be obtained by crossing the rotated local z-axis with the local x-axis
    l_y_rot = np.cross(l_z_rot, l_x)
    # Finally, return the normalized local y-axis
    return l_y_rot / np.linalg.norm(l_y_rot)


_link_tables = {
    "Linear              " : "LINK PROPERTY DEFINITIONS 02 - LINEAR",
    "??                  " : "LINK PROPERTY DEFINITIONS 03 - MULTILINEAR",
    "Damper - Exponential" : "LINK PROPERTY DEFINITIONS 04 - DAMPER",
    "???                 " : "LINK PROPERTY DEFINITIONS 05 - GAP",
    "????                " : "LINK PROPERTY DEFINITIONS 06 - HOOK",
    "?????               " : "LINK PROPERTY DEFINITIONS 07 - RUBBER ISOLATOR",
    "??????              " : "LINK PROPERTY DEFINITIONS 08 - SLIDING ISOLATOR",
    "Plastic (Wen)"        : "LINK PROPERTY DEFINITIONS 10 - PLASTIC (WEN)",
    "???????             " : "LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC",
}

def create_links(csi, model, library, config, conv):

    for link in csi.get("CONNECTIVITY - LINK",[]):

        nodes = (
            conv.identify("Joint", "node", link["JointI"]),
            conv.identify("Joint", "node", link["JointJ"])
        )

        # if any(not (isinstance(node, int) or node.isdigit()) for node in nodes):
        #     conv.log(UnimplementedInstance(f"Joint: non-integer nodes", assign))
        #     continue

        assign = find_row(csi["LINK PROPERTY ASSIGNMENTS"],
                          Link=link["Link"])

        if assign.get("LinkJoints", "") == "SingleJoint":

            props = find_row(csi["LINK PROPERTY DEFINITIONS 01 - GENERAL"],
                             Link=assign["LinkProp"])

            if props["LinkType"] != "Linear":
                conv.log(UnimplementedInstance(f"Joint.SingleJoint.LinkType={props['LinkType']}", assign))

            # TODO: Implement soil springs
            props = find_rows(csi["LINK PROPERTY DEFINITIONS 02 - LINEAR"],
                             Link=assign["LinkProp"])

            flags = tuple(
                1 if find_row(props, DOF=f"{dof[0]}{'XYZ'.find(dof[1])+1}") and config["dofs"][dof]
                  else 0
                  for dof in config["dofs"]
            )

            model.fix(nodes[0], flags)

            continue

        elif assign.get("LinkJoints", "") != "TwoJoint":
            conv.log(UnimplementedInstance(f"Joint.{assign.get('LinkJoints','')}", assign))
            continue

        #
        # Get mats and dofs
        #
        mats = tuple(library["link_materials"][assign["LinkProp"]].values())
        dofs = tuple(library["link_materials"][assign["LinkProp"]].keys())
        dofs = tuple(["U1", "U2", "U3", "R1", "R2", "R3"].index(i)+1 for i in dofs)

        if len(dofs) == 0:
            conv.log(UnimplementedInstance(f"Joint.DOFS", assign))
            continue

        # Check whether there are any zero-length link elements
        xi = np.array(model.nodeCoord(nodes[0]))
        xj = np.array(model.nodeCoord(nodes[1]))

        distance = np.linalg.norm(xj - xi)
        zero_length_threshold = 1e-6

        #
        # Get axes and orientation
        #
        axes = find_row(csi.get("LINK LOCAL AXES ASSIGNMENTS 1 - TYPICAL", []),
                        Link=link["Link"])

        orient_vector = None  # Default value

        if axes:
            if axes["AdvanceAxes"]:
                # Handle advanced axes
                axes_advance = find_row(csi.get("LINK LOCAL AXES ASSIGNMENTS 2 - ADVANCED", []),
                                        Link=link["Link"])

                # Common advanced axes setup
                orient_vector = (
                    axes_advance["AxVecX"], axes_advance["AxVecY"], axes_advance["AxVecZ"],
                    axes_advance["PlVecX"], axes_advance["PlVecY"], axes_advance["PlVecZ"],
                )

                # Additional validation for non-zero-length links
                if distance > zero_length_threshold:
                    # Calculate node-based orientation
                    orient_vector_from_nodes = _orient(xi, xj, axes["Angle"])

                    # Calculate y-axis from advanced axes
                    ax_vec = np.array([axes_advance["AxVecX"], axes_advance["AxVecY"], axes_advance["AxVecZ"]])
                    pl_vec = np.array([axes_advance["PlVecX"], axes_advance["PlVecY"], axes_advance["PlVecZ"]])

                    x_axis = ax_vec / np.linalg.norm(ax_vec)
                    pl_projection = pl_vec - np.dot(pl_vec, x_axis) * x_axis
                    y_axis_advanced = pl_projection / np.linalg.norm(pl_projection)

                    # Validate and update orientation
                    if np.allclose(y_axis_advanced, orient_vector_from_nodes, atol=1e-6):
                        orient_vector = tuple(_orient(xi, xj, axes["Angle"]))
                    else:
                        orient_vector = tuple(_orient(xi, xj, axes["Angle"]))
                        warnings.warn(f"Orientation mismatch in link {link['Link']}")
#                       raise ValueError(f"Orientation mismatch in link {link['Link']}")

            else:
                # Handle typical axes (non-advanced)
                orient_vector = tuple(_orient(xi, xj, axes["Angle"]))


        #
        # Create the link
        #
        tag = conv.define("Link", "element", link["Link"])
        if orient_vector is not None:
            model.element("TwoNodeLink",
                      tag,
                      nodes,
                      mat=mats,
                      dir=dofs,
                      orient=orient_vector
                      )
        else:
            model.element("TwoNodeLink",
                      tag,
                      nodes,
                      mat=mats,
                      dir=dofs
                      )

    return



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
from xcsi.csi import create_model
from xcsi.names import Names
from dataclasses import dataclass
from xcsi.csi.utility import find_row



class Assembly:
    def __init__(self, tree, model=None):
        self._tree = tree
        self._names = Names()
        self._model = create_model(tree, model=model, names=self._names, verbose=False)

    def frame_shape(self, name=None, tag=None):
        pass

    @property
    def units(self):
    
        units = None 
        unit_str = None
        for row in self._tree["PROGRAM CONTROL"]:
            if "CurrUnits" in row:
                unit_str = row["CurrUnits"]
        
        if unit_str.lower() == "kip, ft, f":
            import xara.units.fks as units
        elif unit_str.lower() == "kip, in, f":
            import xara.units.iks as units
        elif unit_str.lower() == "lb, in, f":
            import xara.units.ips as units
        elif unit_str.lower() == "kn, m, c":
            import xara.units.mks as units

        return units

    @property 
    def names(self):
        return self._names

    @property 
    def model(self):
        return self._model


class Job:
    def __init__(self, tree):
        self._tree = tree

    def assemble(self, model=None):
        return Assembly(self._tree, model=model)
    
    def pattern(self, name):
        from xcsi.csi.analysis import LoadPattern
        return LoadPattern(name, self._tree)
    
    def patterns(self):
        for pattern in self._tree.get("LOAD PATTERN DEFINITIONS", []):
            yield self.pattern(pattern["LoadPat"])
    
    def steps(self):
        from xcsi.csi.analysis import LinearStaticAnalysis

        for case in self._tree.get("LOAD CASE DEFINITIONS", []):
            if case["Type"] == "LinStatic":
                yield LinearStaticAnalysis(case["Case"], self)
            else:
                raise ValueError(f"Unsupported case type: {case['Type']}")



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import warnings 
from .utility import find_row, find_rows
import numpy as np


class _Section:
    def __init__(self, name: str, csi: dict,
                 index: int, model, library, conv):
        self.index = index
        self.name = name
        self.integration = []

        self._create(csi, model, library, conv)

    def _create(self, csi, model, library, conv):
        pass


def _create_shell_integration(csi, model, conv):
    pass


def _create_shell_section(csi, assign, model, conv):

    section = find_row(csi["AREA SECTION PROPERTIES"],
                        Section=assign["Section"]
    )

    # assert section is not None, f"Section {assign['Section']} not found in AREA SECTION PROPERTIES"
    if section is None:
        # TODO: log
        print(assign["Section"])

    tag = conv.define("ShellSection", "section", assign["Section"])

    material = find_row(csi["MATERIAL PROPERTIES 01 - GENERAL"],
                        Material=section["Material"]
    )

    material = find_row(csi["MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES"],
                        Material=section["Material"]
    )

    model.section("ElasticShell", tag,
                    material["E1"],  # E
                    material["G12"]/(2*material["E1"]) - 1, # nu
                    section["Thickness"],
                    material["UnitMass"]
    )
    # self.integration.append(self.index)
    return tag


def add_shell_sections(csi, model, conv):
    for assign in csi.get("AREA SECTION ASSIGNMENTS", []):
        # if "Section" not in assign:
        #     print(assign)
        #     continue
        if not conv.identify("ShellSection", "section",  assign["Section"]):

            if not _create_shell_section(csi, assign, model, conv):
                warnings.warn(f"Section {assign['Section']} not found")
                continue

            # tag += len(library["shell_sections"][assign["Section"]].integration)




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
from xcsi.csi import create_model, load, collect_outlines, find_rows
from xcsi.job import Job
from opensees.repl.ptkshell import OpenSeesREPL



def main():
    import sys

    file_name = None
    out_file = None
    operation = "View"
    strict_mode = None
    object_mode = "Model"
    interact = False
    show = set()
    objects = []

    ic = None
    i  = 0
    argi = iter(sys.argv[1:])
    for arg in argi:
        if arg == "-s":
            object_mode = "Section"
        elif arg == "-t":
            object_mode = next(argi)
            if object_mode not in ["Model", "Section", "Pattern", "Step"]:
                raise ValueError(f"Unknown object mode {object_mode}")

        elif arg == "-o":
            try:
                out_file = next(argi)
            except StopIteration:
                raise ValueError("Flag -o missing required argument <file>")

        elif arg == "-i":
            interact = True

        elif arg == "-m":
            try:
                strict_mode = next(argi)
            except StopIteration:
                raise ValueError("Flag -m missing required argument <mode>")
        
        elif arg == "--show":
            try:
                show.add(next(argi))
            except StopIteration:
                raise ValueError("Flag --show missing required argument <name>")

        elif arg[0] == "-":
            operation = {
                "p": "Print",
                "a": "Analyze",
                "c": "Convert",
                "x": "Execute",
                "v": "Visualize",
            }[arg[1].lower()]
            ic = i+1

        elif file_name is None:
            file_name = arg
        else:
            objects.append(arg.split("="))

        i += 1

    if file_name is None:
        raise ValueError("No file specified")
    

    with open(file_name, "r") as f:
        csi = load(f)


    if operation == "Convert":
        # Convert
        if "tcl" in sys.argv[1]:
            import xara 
            model = xara.Model(ndm=3, ndf=6, echo_file=sys.stdout)
            model = create_model(csi, model=model, verbose=False)

        else:
            model = create_model(csi, verbose=False)
            model.print("-json")
        sys.exit()

    
    if operation == "Print":
        job = Job(csi)
        if object_mode == "Pattern":
            for pattern in job.patterns():
                print(pattern)
        elif object_mode == "Step":
            for step in job.steps():
                print(step)
        elif len(objects) == 0:
            for table in csi:
                print(f"{table}: {len(csi[table])}")
        else:
            for key, value in objects:
                for table in csi:
                    filt = {key: value}
                    if (row:= find_rows(csi[table], **filt)):
                        print(table)
                        for r in row:
                            for k, v in r.items():
                                if k == key:
                                    continue
                                print(f"  {k:<10}      {v}")
                            print()
                            continue
        sys.exit()



    if operation == "Execute":
        job = Job(csi)

        asm = job.assemble()

        for step in job.steps():
            step.run(asm)
            break

        if interact:
        
            # Start an interactive console
            repl = OpenSeesREPL(asm.model._openseespy._interp)
            repl.repl()


    elif operation == "Visualize":

        # Visualize
        import veux
        model = create_model(csi, verbose=True)
        outlines = collect_outlines(csi, model.frame_tags)
        artist = veux.create_artist(model, canvas="gltf", vertical=3,
                    model_config={
                        "frame_outlines": outlines
                    }
        )
        # artist.draw_nodes()
        artist.draw_outlines()
        artist.draw_surfaces()


        if sys.argv[1] == "-Vo":
            artist.save(sys.argv[3])
        else:
            veux.serve(artist)

    elif sys.argv[1] == "-Vn":
        # Visualize
        from scipy.linalg import null_space
        model.constraints("Transformation")
        model.analysis("Static")
        K = model.getTangent().T
        v = null_space(K)[:,0] #, rcond=1e-8)
        print(v)


        u = {
            tag: [1000*v[dof-1] for dof in model.nodeDOFs(tag)]
            for tag in model.getNodeTags()
        }

        import veux
        veux.serve(veux.render(model, u, canvas="gltf", vertical=3))

    elif sys.argv[1] == "-Q":
        # Quiet conversion
        pass
    else:
        raise ValueError(f"Unknown operation {sys.argv[1]}")


if __name__ == "__main__":
    main()


"""
    MIT License
    
    Copyright (c) 2020 OpenSeesPro
 
    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.
    
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    Developed by:
        Ayush Singhania (ayushs@stanford.edu)
        Pearl Ranchal (ranchal@berkeley.edu)
      
    Publication:
        Goings, C. B., Singhania, A., Ranchal, P., Weaver B., 2020, “Industrial 
        Scale NLRH Analysis Using OpenSees and Comparison with Perform3D,” 
        Proceedings of 2020 SEAOC Virtual Convention, SEAOC, CA
    
    Description of the script - 
        This is a supporting script for the main.py (user should execute main.py)
        This script is a utility script to perform tasks supporting scraping data
        from ETABS as well as creating model in OpenSees and running analysis.

"""

import comtypes.client
import sys
import time
import pandas as pd
import numpy as np


def get_model_from_etabs():
    try:
        # attach to a running instance of ETABS
        etabs = comtypes.client.GetActiveObject("CSI.ETABS.API.ETABSObject")
    except (OSError, comtypes.COMError):
        print("No running instance of the program found or failed to attach.")
        sys.exit(-1)
    return etabs.SapModel


def set_load_cases_selected_for_display(loadCaseList, model=get_model_from_etabs()):
    return model.DatabaseTables.SetLoadCasesSelectedForDisplay(loadCaseList)


def set_load_combo_selected_for_display(loadComboList, model=get_model_from_etabs()):
    return model.DatabaseTables.SetLoadCombinationsSelectedForDisplay(loadComboList)


def set_load_patterns_selected_for_display(
    loadPatternList, model=get_model_from_etabs()
):
    return model.DatabaseTables.SetLoadPatternsSelectedForDisplay(loadPatternList)


def deselect_all_load_cases_and_combos_for_output(model=get_model_from_etabs()):
    return [
        bool(~set_load_cases_selected_for_display("", model)[-1]),
        bool(~set_load_combo_selected_for_display("", model)[-1]),
    ]


def get_database_table_for_all_load_cases_and_combos(table_title, model):

    # set units to kip, in (default for OpenSees)
    model.SetPresentUnits(3)

    # get lists of all load cases and combinations
    try:
        listOfLoadCases = [
            i for i in list(model.LoadCases.GetNameList()[-2]) if "~" not in i
        ]
    except:
        listOfLoadCases = []

    try:
        listOfLoadCombos = list(model.RespCombo.GetNameList()[-2])
    except:
        listOfLoadCombos = []

    # select all load cases and combinations for output
    model.DatabaseTables.SetLoadCombinationsSelectedForDisplay(listOfLoadCombos)
    model.DatabaseTables.SetLoadCasesSelectedForDisplay(listOfLoadCases)

    # get table from API
    ret = model.DatabaseTables.GetTableForDisplayArray(table_title, "", "")

    # develop DataFrame
    headers = list(ret[2])
    data = np.array(ret[4])
    data = data.reshape(len(data) // len(headers), len(headers))
    df = pd.DataFrame(data)

    cols_dict = dict(zip(df.columns.tolist(), headers))
    df.rename(cols_dict, axis="columns", inplace=True)

    return df.copy()


def start_time():
    print("Started at: " + time.strftime("%a, %d %b %Y %H:%M:%S PST", time.localtime()))
    return time.time()


def end_time(start_time, final=True):
    end_time = time.time()
    (mins, secs) = divmod(round(end_time - start_time, 0), 60)
    print(f"\nTime Elapsed: {mins} minute(s) {secs} second(s).\n")
    if final:
        print(
            "Finished at: "
            + time.strftime("%a, %d %b %Y %H:%M:%S PST", time.localtime())
        )
    return



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# April 29, 2025
#
from .utility import UnimplementedInstance, find_row
from ..names import RE, Names

def add_solids(csi, model, config, names: Names):
    ndm = config.get("ndm", 3)
    log = []


    for elem in csi.get("CONNECTIVITY - SOLID",[]):
        nodes = tuple(
                names.identify("Joint", "node", v) or print(v)
                for k,v in elem.items() if RE["joint_key"].match(k)
        )

        nodes = (
            nodes[0],
            nodes[1],
            nodes[3],
            nodes[2],
            nodes[4],
            nodes[5],
            nodes[7],
            nodes[6]
        )

        props = find_row(csi.get("SOLID PROPERTY ASSIGNMENTS", []), Solid=elem["Solid"])
        if props is None:
            log.append(UnimplementedInstance("Solid", elem))
            continue
        props = find_row(csi.get("SOLID PROPERTY DEFINITIONS", []), SolidProp=props["SolidProp"])
        if props is None:
            log.append(UnimplementedInstance("Solid", elem))
            continue

        material = names.identify("Material", "material", props["Material"])
        if material is None:
            raise ValueError(f"Material {props['Material']} not found")

        if len(nodes) == 8:
            model.element("stdBrick", 
                          names.define("Solid", "element", elem["Solid"]),
                          nodes,
                          material
            )
        else:
            log.append(UnimplementedInstance("Solid", elem))
            continue


#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Certain operations are loosley adapted from:
#    https://github.com/XunXun-Zhou/Sap2OpenSees/blob/main/STO_ver1.0.py
#
#
import numpy as np
from ..names import Names
from .parse import load
from .utility import UnimplementedInstance, print_log
from .point import create_points
from .link import create_links
from ._section import add_shell_sections
from ._frame.section import add_frame_sections
from ._frame.outlines import collect_geometry as collect_outlines
from .utility import find_row, find_rows

CONFIG = {
    "Frame": {
        "Taper": "Subdivide", # Integrate
        "Element": "PrismFrame",
    }
}



def create_materials(csi, model, conv):
    "SD STRESS-STRAIN 01 - REBAR PARK"
    "SD STRESS-STRAIN 02 - REBAR SIMPLE"
    "SD STRESS-STRAIN 03 - STRUCTURAL STEEL"
    "SD STRESS-STRAIN 04 - CONCRETE SIMPLE"
    "SD STRESS-STRAIN 05 - CONCRETE MANDER UNCONFINED"
    "SD STRESS-STRAIN 06 - CONCRETE MANDER CONFINED CIRCLE"
    "SD STRESS-STRAIN 07 - CONCRETE MANDER CONFINED RECTANGLE"
    library = conv._library

    # 1) Material

    #
    # 2) Links
    #
    mat_total = 1

    for mat in csi.get("MATERIAL PROPERTIES 01 - GENERAL", []):
        if mat["SymType"] == "Isotropic":
            pass
        else:
            conv.log(UnimplementedInstance("Material", mat))
            continue
        
        p02 = find_row(csi.get("MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES", []), Material=mat["Material"])

        name = mat["Material"]
        model.nDMaterial(
            "ElasticIsotropic",
            conv.define("Material", "material", name),
            p02["E1"],  # Young
            p02["U12"], # Poisson
        )

    for link in csi.get("LINK PROPERTY DEFINITIONS 02 - LINEAR", []):
        if link["Fixed"]:
            conv.log(UnimplementedInstance("Link.Fixed", link))
            continue

        name = link["Link"]
        if "R" in link["DOF"]:
            stiff = link["RotKE"]
            damp  = link["RotCE"]
        else:
            stiff = link["TransKE"]
            damp  = link["TransCE"]

        # TODO: use damp
        model.eval(f"material ElasticIsotropic {mat_total} {stiff} 0.3\n")

        dof = link["DOF"]
        library["link_materials"][name][dof] = mat_total
        mat_total += 1

    for damper in csi.get("LINK PROPERTY DEFINITIONS 04 - DAMPER", []):
        # TODO: implement dampers
        conv.log(UnimplementedInstance("Link.Damper", damper))
        continue
        name = damper["Link"]
        stiff = damper["TransK"]
        dampcoeff = damper["TransC"]
        exp = damper["CExp"]
        model.eval(f"uniaxialMaterial ViscousDamper {mat_total} {stiff} {dampcoeff}' {exp}\n")

        dof = damper["DOF"]
        library["link_materials"][name][dof] = mat_total
        mat_total += 1


    for link in csi.get("LINK PROPERTY DEFINITIONS 10 - PLASTIC (WEN)", []):
        name = link["Link"]

        if not link.get("Nonlinear", False):
            stiff = link["TransKE"]
            model.eval(f"uniaxialMaterial Elastic {mat_total} {stiff}\n")
        else:
            stiff = link["TransK"]
            fy    = link["TransYield"]
            exp   = link["YieldExp"] # TODO
            ratio = link["Ratio"]
            model.eval(f"uniaxialMaterial Steel01 {mat_total} {fy} {stiff} {ratio}\n")

        dof = link["DOF"]
        library["link_materials"][name][dof] = mat_total
        mat_total += 1
    return library


def _create_sections(csi, model, conv):
    # 2) Frame
    if len(csi.get("CONNECTIVITY - FRAME", [])) > 0:
        add_frame_sections(csi, model, conv)


    # 3) Shell
    add_shell_sections(csi, model, conv)



def assemble(csi, model=None, verbose=False):
    from xcsi.job import Job
    return Job(csi).assemble(verbose=verbose)



def create_model(csi, types=None, model=None, verbose=False, names=None):
    """
    Parameters
    ==========
    csi: a dictionary formed by calling ``csi.parse.load("file.b2k")``

    Returns
    =======
    model: opensees.openseespy.Model object
    """

    import opensees.openseespy as ops

    config = CONFIG

    used = {
        "TABLES AUTOMATICALLY SAVED AFTER ANALYSIS"
    }


    #
    # Create model
    #
    dofs = {key:val for key,val in csi["ACTIVE DEGREES OF FREEDOM"][0].items() } # if val }
    dims = {key for key,val in csi["ACTIVE DEGREES OF FREEDOM"][0].items() } # if val }
    ndf = sum(1 for v in csi["ACTIVE DEGREES OF FREEDOM"][0].values())
    ndm = sum(1 for k,v in csi["ACTIVE DEGREES OF FREEDOM"][0].items()
              if k[0] == "U")

    if isinstance(verbose, int) and verbose > 3:
        import sys
        echo_file = sys.stdout
    else:
        echo_file = None

    if model is None:
        model = ops.Model(ndm=ndm, ndf=ndf, echo_file=echo_file)

    if names is None:
        names = Names()

    used.add("ACTIVE DEGREES OF FREEDOM")

#   dofs = [f"U{i}" for i in range(1, ndm+1)]
#   if ndm == 3:
#       dofs = dofs + ["R1", "R2", "R3"]
#   else:
#       dofs = dofs + ["R3"]

    config["ndm"]  = ndm
    config["ndf"]  = ndf
    config["dofs"] = dofs

    #
    # Create nodes
    #
    create_points(csi, model, None, config, names)

    #
    # Create materials and sections
    #
    library = create_materials(csi, model, names)

    _create_sections(csi, model, names)


    # Unimplemented objects
    for item in [
        "CONNECTIVITY - CABLE",
        "CONNECTIVITY - TENDON"]:
        for elem in csi.get(item, []):
            names.log(UnimplementedInstance(item, elem))

    #
    # Add elements
    #
    _add_elements(model, csi, library, config, names)


    if verbose and len(names._log) > 0:
        print_log(names._log)

    if verbose and False:
        for table in csi:
            if table not in used:
                print(f"\t{table}", file=sys.stderr)


    model.frame_tags = library.get("frame_tags", {})
    return model


def _add_elements(model, csi, library, config, names):
    """
    Add elements to the model.
    """
    from ._frame import add_frames
    from ._shell import add_shells
    from ._solid import add_solids

    create_links(csi, model, library, config, names)
    add_frames(csi, model, library, config, names)
    add_shells(csi, model, names)
    add_solids(csi, model, config, names)



from xsection.library import (
    WideFlange,
    Rectangle,
    HollowRectangle,
    Angle,
    Channel,
    HollowRectangle,
    from_aisc
)

if False:
    def from_aisc(identifier: str, mesh:tuple=None, units=None, ndm=None, tag=None, **kwds):
        if units is None:
            import opensees.units.english as _units
            units = _units

        aisc_data = load_aisc(identifier, units=units)

        if aisc_data is None:
            raise ValueError(f"Cannot find section with identifier {identifier}")


        if identifier[0] == "W":
            return WideFlange(d=aisc_data["d"],
                            b=aisc_data["bf"],
                            t=aisc_data["tf"],
                            tw=aisc_data["tw"],
                            k=aisc_data.get("k1", None),
                            **kwds)

        elif identifier[0] == "L":
            return Angle(d=aisc_data["d"],
                        b=aisc_data["b2"],
                        t=aisc_data["t"])

        elif identifier[0] == "C":
            return Channel(d=aisc_data["d"],
                        b=aisc_data["bf"],
                        tf=aisc_data["tf"],
                        tw=aisc_data["tw"])
        elif identifier[:3] == "HSS":
            return HollowRectangle(d=aisc_data["h/tdes"]*aisc_data["tdes"],
                                b=aisc_data["b/tdes"]*aisc_data["tdes"],
                                t=aisc_data["tdes"],
            )
        else:
            raise ValueError(f"Cannot create section from identifier {identifier}")



    def load_aisc(SectionName, props="", units=None)->dict:
        """Load cross section properties from AISC database.

        props:
            A list of AISC properties, or one of the following:
            - 'simple': `A`, `Ix`, `Zx`

        """

        if units is None:
            import opensees.units.english as _units
            units = _units

        from shps.frame.shapes.aisc_imperial import imperial
        SectData = imperial[SectionName.upper()]


        if props == "simple":
            props = ""
            return

        elif props:
            props = props.replace(" ", "").split(",")
            sectData = {k: v for k, v in SectData.items() if k in props}
            if "I" in props:
                sectData.update({"I": SectData["Ix"]})
            return sectData

        for k,v in list(SectData.items()):
            try:
                SectData[k] = float(v)
            except:
                continue


        UNITS = [
            ("d"  ,  units.inch ),
            ("k1"  , units.inch ),
            ("h/tdes", 1 ),
            ("b/tdes", 1 ),
            ("tdes", units.inch ),
            ("bf" , units.inch ),
            ("tw" , units.inch ),
            ("tf" , units.inch ),
            ("A"  , units.inch**2 ),
            ("Ix" , units.inch**4 ),
            ("Iy" , units.inch**4 ),
        ]

        return {
            k: SectData[k]*scale
            for k, scale in UNITS if k in SectData and isinstance(SectData[k], (float, int))
        }



if __name__ == "__main__":
    import veux
    d  = 100
    tw = 3
    bf = 75
    tf = 3

    mesh = WideFlange(d=d, b=bf, t=tf, tw=tw).create_shape()

    print(mesh.summary())

#   from shps.frame.solvers.plastic import PlasticLocus
#   PlasticLocus(mesh).plot()#(phi=0.5, ip=5)
#   import matplotlib.pyplot as plt
#   plt.show()

    artist = veux.create_artist(((mesh.mesh.nodes, mesh.mesh.cells())), ndf=1)

    field = mesh.torsion.warping()
    field = {node: value for node, value in enumerate(field)}

    artist.draw_surfaces(field = field)
    artist.draw_origin()
#   R = artist._plot_rotation
#   artist.canvas.plot_vectors([R@[*geometry.centroid, 0] for i in range(3)], R.T)
    artist.draw_outlines()
    veux.serve(artist)



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import json
import shlex

CONSTANTS = {
        "Yes": True,
        "No":  False
}

def _parse_value(v):
    if v in CONSTANTS:
        return CONSTANTS[v]

    try:
        return json.loads(v)
    except:
        return v


def load(file, append: dict=None):
    """
    Read file-like object file and form a dictionary. 

    Returns
    =======

    csi: A dictionary with the layout:
        {
           "TABLE NAME": [
            {"Key": "Val"},
            ...
           ]
        }
    """
    if append is None:
        tables = {}
    else:
        tables = append

    current_table = None
    current_item  = None
    for line_no, line in enumerate(file):
        if "END TABLE DATA" in line:
            break

        # Skip empty lines
        if line.isspace():
            continue

        if "TABLE:" in line:
            table_name = shlex.split(line)[1]
            current_item  = None

            # Append if table exists (append argument given)
            if table_name in tables:
                current_table = tables[table_name]
            else:
                current_table = []

            tables[table_name] = current_table


        # Data line
        elif current_table is not None:

            if current_item is None:

                current_item = {}
                current_table.append(current_item)

            continue_item = False
            try:
                # Things are complicated by the fact we have to parse things like:
                #    Key=Val'
                # into "Key": "Val'"
                #
                lex = shlex.shlex(line, posix=True)
                lex.quotes = '"'
                lex.wordchars += "'"
                lex.commenters = ""
                lex.whitespace_split = True
                tokens = list(lex)
            except Exception as e:
                raise ValueError(f"Error parsing line {line_no}: " + str(e))

            for i,kv in enumerate(tokens):
                # A "_" usually means a continuation of the previous line
                # but sometimes it is just a random character.
                if kv == "_":
                    if i == len(tokens)-1:
                        continue_item = True
                        break

                    # Sometimes there is a random "_" in the middle
                    # of a line?
                    else:
                        continue

                k, v = kv.split("=", maxsplit=1)
                current_item[k] = _parse_value(v)

            if not continue_item:
                current_item = None

    return tables




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#

class PeakDrift:
    def __init__(self, joints):
        self._joints = joints


    def record(self, asm):
        model = asm.model

        i = asm.names.identify("Joint", "node", self._joints[0])
        j = asm.names.identify("Joint", "node", self._joints[1])

        model.recorder("Node", "disp", node=(i,j), file=str(id(self)), dof=(1,2,3))


    def draw(self):
        pass


    def retrieve(self):
        return



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
from ..utility import find_row, find_rows
import numpy as np
import warnings
from veux.frame import SectionGeometry

from .section import CIRCLE_DIVS as _CIRCLE_DIVS, _CsiSection


def _find_polygons(csi, prop_01) -> SectionGeometry:
    """
    was previously section_geometry(csi, prop_01)

    POST-PROCESSING
    """
    #
    if isinstance(prop_01, str):
        name = prop_01
        prop_01 = find_row(csi.get("FRAME SECTION PROPERTIES 01 - GENERAL",[]), SectionName=name)
        if prop_01 is None:
            prop_01 = find_row(csi.get("FRAME SECTION PROPERTIES - BRIDGE OBJECT FLAGS",[]), SectionName=name)
            if prop_01 is None:
                raise ValueError(f"Section {name} not found in either table.")
    else:
        name = prop_01["SectionName"]

    s = _CsiSection(csi, prop_01)._create_model(render_only=True)
    if s is not None:
        return SectionGeometry(s.exterior(), interior=s.interior())


# def _section_mesh(csi, prop_01, engine=None):
#     """
#     """
#     from shps.frame.mesh import sect2meshpy
#     shape = _find_polygons(csi, prop_01)
#     if engine is None:
#         shape = (
#             shape.exterior(plane=True),
#             shape.interior(plane=True)
#         )
#         return sect2meshpy(shape, 0.5)

#     geometry = _find_polygons(csi, prop_01)
#     exterior = geometry.exterior(plane=True)
#     interior = geometry.interior(plane=True)

#     import gmsh
#     import meshio
#     gmsh.initialize()
#     gmsh.model.add("section")
#     # Add exterior points
#     exterior_points = [gmsh.model.geo.addPoint(x, y, 0) for x, y in exterior]
#     exterior_loop = gmsh.model.geo.addCurveLoop([gmsh.model.geo.addSpline(exterior_points)])

#     # Add interior points (holes)
#     interior_loops = []
#     for hole in interior:
#         hole_points = [gmsh.model.geo.addPoint(x, y, 0) for x, y in hole]
#         interior_loops.append(gmsh.model.geo.addCurveLoop([gmsh.model.geo.addSpline(hole_points)]))

#     # Create plane surface with holes
#     surface = gmsh.model.geo.addPlaneSurface([exterior_loop] + interior_loops)

#     # Synchronize to create the surface
#     gmsh.model.geo.synchronize()

#     # Generate 2D mesh
#     gmsh.model.mesh.generate(2)
    
#     # Extract mesh data
#     nodes = gmsh.model.mesh.getNodes()
#     elements = gmsh.model.mesh.getElements()
    
#     # Convert to meshio format
#     mesh = meshio.Mesh(
#         points=nodes[1].reshape(-1, 3)[:, :2],  # Only take x, y coordinates
#         cells={"triangle": elements[2][0].reshape(-1, 3) - 1}  # Convert to 0-based indexing
#     )
    
#     # Finalize gmsh
#     gmsh.finalize()
    
#     return mesh


class FrameQuadrature:
    def __init__(self, sections, locations=None, geometry: list=None, name=None):
        self._sections  = sections
        self._locations = locations
        self._geometry  = geometry
        self._name = name 

    def __repr__(self):
        if self._name is not None:
            return f"FrameQuadrature({self._name})"
        else:
            return f"FrameQuadrature({len(self._sections)} sections, {len(self._geometry)} geometries)"

    @classmethod
    def from_table(cls, csi, prop_01):
        # 1)
        name = prop_01["SectionName"]
        if prop_01["Shape"] != "Nonprismatic":
            geometry = _find_polygons(csi, prop_01)
            if geometry is not None:
                geometry = [geometry, geometry]

            # section = _FrameSection.from_table(csi, prop_01)

            return FrameQuadrature([], #[section, section],
                                   geometry=geometry, name=name)

        row = find_row(csi.get("FRAME SECTION PROPERTIES 05 - NONPRISMATIC", []),
                        SectionName=prop_01["SectionName"])

        # 2)
        if row["StartSect"] == row["EndSect"]:
            si = find_row(csi["FRAME SECTION PROPERTIES 01 - GENERAL"], 
                                SectionName=row["StartSect"])

            assert si is not None

            geometry = _find_polygons(csi, si)
            if geometry is not None:
                geometry = [geometry, geometry]
            # section = _FrameSection.from_table(csi, prop_01)
            return FrameQuadrature([], #[section, section],
                                   geometry=geometry, name=name)

        # 3)
        else:
            si = find_row(csi["FRAME SECTION PROPERTIES 01 - GENERAL"],
                            SectionName=row["StartSect"])
            sj   = find_row(csi["FRAME SECTION PROPERTIES 01 - GENERAL"],
                            SectionName=row["EndSect"])

            if si["Shape"] == sj["Shape"] and si["Shape"] in {"Circle"}:
                circumference = np.linspace(0, np.pi*2, _CIRCLE_DIVS)
                exteriors = np.array([
                    [[np.sin(x)*r, np.cos(x)*r] for x in circumference]
                    for r in np.linspace(si["t3"]/2, sj["t3"]/2, 2)
                ])
                return FrameQuadrature([], 
                                       geometry = [SectionGeometry(exterior) for exterior in exteriors])

    # def sections(self):
    #     return self._sections

    def locations(self):
        pass 

    def geometry(self):
        if self._geometry is not None:
            return [i for i in self._geometry]

    def weights(self):
        pass


def collect_geometry(csi, elem_maps=None):
    """
    collect a mapping of SectionGeometry for all frames elements in a model.

    This is a POST-PROCESSING function, used for rendering extruded frames.

    TODO: Reimplement as loop over iter_sections()
    """

    frame_types = {
        row["SectionName"]: FrameQuadrature.from_table(csi, row)
        for row in csi.get("FRAME SECTION PROPERTIES 01 - GENERAL", [])
    }

    frame_assigns = {}
    for row in csi.get("FRAME SECTION ASSIGNMENTS",[]):
        if row["MatProp"] != "Default":
            warnings.warn(f"Material property {row['MatProp']} not implemented.")

        if row["AnalSect"] in frame_types and frame_types[row["AnalSect"]] is not None:
            if frame_types[row["AnalSect"]].geometry() is None:
                warnings.warn(f"No geometry for {row['AnalSect']}")
                continue
            frame_assigns[row["Frame"]] = frame_types[row["AnalSect"]].geometry()
            

    # Skew angles
    E2 = np.array([0, 0,  1])
    for frame in frame_assigns:
        skew_assign = find_row(csi.get("FRAME END SKEW ANGLE ASSIGNMENTS", []),
                               Frame=frame)
        
        if skew_assign: #and skew["SkewI"] != 0 and skew["SkewJ"] != 0: # and len(frame_assigns[frame].shape) == 2
            for i,skew in zip((0,-1), ("SkewI", "SkewJ")):
                exterior = frame_assigns[frame][i].exterior()
                interior = frame_assigns[frame][i].interior()

                R = _ExpSO3(skew_assign[skew]*np.pi/180*E2)
                frame_assigns[frame][i] = SectionGeometry(interior=[np.array([[(R@point)[0], *point[1:]] for point in hole]) for hole in interior],
                                                          exterior=np.array([[(R@point)[0], *point[1:]]  for point in exterior])
                )

    if elem_maps is not None:
        return {
            elem_maps.get(name,name): val for name, val in frame_assigns.items()
        }
    else:
        return frame_assigns


def _HatSO3(vec):
    """Construct a skew-symmetric matrix from a 3-vector."""
    return np.array([
        [0, -vec[2], vec[1]],
        [vec[2], 0, -vec[0]],
        [-vec[1], vec[0], 0]
    ])

def _ExpSO3(vec):
    """
    Exponential map for SO(3).
    Satisfies ExpSO3(vec) == expm(skew(vec)).
    """
    vec = np.asarray(vec)
    if vec.shape != (3,):
        raise ValueError("Input must be a 3-vector.")

    theta = np.linalg.norm(vec)
    if theta < 1e-8:  # Small-angle approximation
        return np.eye(3) + _HatSO3(vec) + 0.5 * (_HatSO3(vec) @ _HatSO3(vec))
    else:
        K = _HatSO3(vec / theta)  # Normalized skew matrix
        return np.eye(3) + np.sin(theta) * K + (1 - np.cos(theta)) * (K @ K)




"""
MIT License

Developed by:
    Ayush Singhania (ayushs@stanford.edu)
    Pearl Ranchal (ranchal@berkeley.edu)
  
Publication:
    Goings, C. B., Singhania, A., Ranchal, P., Weaver B., 2020, “Industrial 
    Scale NLRH Analysis Using OpenSees and Comparison with Perform3D,” 
    Proceedings of 2020 SEAOC Virtual Convention, SEAOC, CA

Description of the script - 
    This is a supporting script for the main.py (user should execute main.py)
    This script is used to generate the OpenSees model and execute analyses
    in OpenSees.

"""

import opensees.openseespy as op
import opensees.postprocessing.Get_Rendering as opp
import pandas as pd
import numpy as np
import math
import os, shutil
from tqdm import tqdm

# some constants that may be used for OpenSees model creation
g = 386.4
M = 0
E = 29000
G = 11153.846
numEigen = 3
rigid_dia = True
col_transf_tag = 1
beam_transf_tag = 2
coordTransf = "Linear"  # Can be {Linear, PDelta, Corotational}
massType = "-lMass"  # Can be {-lMass, -cMass}
tol = 1e-3

# SETUP OPENSEES MODEL
def setup_opensees_model(
    joints_df,
    frames_df,
    frame_props_df,
    pts_loads_df,
    mass_df,
    dict_of_hinges,
    dict_of_hinges_2,
    list_new_joints,
):

    # set modelbuilder
    model = op.Model(ndm=3)

    add_nodes(model, joints_df.copy(), mass_df.copy(), list_new_joints, dict_of_hinges)
    add_frames(model, frames_df.copy(), frame_props_df.copy())
    add_beam_hinges(model, dict_of_hinges, dict_of_hinges_2)
    return


# ADD NODES TO THE OPENSEES MODEL USING THE DATA FROM ETABS
def add_nodes(op, joints_df, mass_df, list_new_joints, dict_of_hinges):

    # Identify the joints that were auto created by ETABS (like COM joints)
    points_df = joints_df[joints_df.IsAuto == "Yes"].copy()

    # Create joints in opensees mdoel
    joints_df.apply(
        lambda row: op.node(row.UniqueName, row.X, row.Y, row.Z), axis="columns"
    )
    # Add restraints to the joints
    joints_df.apply(
        lambda row: op.fix(row.UniqueName, *[int(r) for r in row.Restraints]),
        axis="columns",
    )
    # Special joint restraints for the COM joints
    points_df.apply(
        lambda row: op.fix(row.UniqueName, *[0, 0, 1, 1, 1, 0]), axis="columns"
    )

    op.constraints("Transformation")

    # If the model diaphragm is rigid, define rigidity on each floor
    if rigid_dia:

        # obtain the list of floors in terms of the elevation (Z coordinate)
        floor_list = set(joints_df[joints_df.Z > joints_df.Z.min()].Z)

        # iterate through each floor to define rigid diaphragm
        for floor in floor_list:
            nodes_list = list(joints_df.loc[joints_df.Z == floor]["UniqueName"])
            remove_list = list(set(nodes_list) & set(list_new_joints)) + list(
                dict_of_hinges.keys()
            )
            nodes = [i for i in nodes_list if i not in remove_list]
            op.rigidDiaphragm(3, *nodes)

    # apply mass to the respective nodes
    mass_df.apply(
        lambda row: op.mass(
            row.PointElm, row.UX, row.UY, row.UZ, row.RX, row.RY, row.RZ
        ),
        axis="columns",
    )

    return


# ADD FRAMES OBJECTS TO THE OPENSEES MODEL
def add_frames(op, frames_df, frame_props_df):

    # function to identify and assign beam and columns their respective tags
    def determine_tag(label):
        if "C" in label:
            tag = col_transf_tag
        else:
            tag = beam_transf_tag
        return tag

    # construct a coordinate-transformation object, which transforms beam and column element
    # stiffness and resisting force from the basic system to the global-coordinate system.
    op.geomTransf(coordTransf, col_transf_tag, 1, 0, 0)
    op.geomTransf(coordTransf, beam_transf_tag, 0, 0, 1)

    # add all the frame elements which are oriented with the major axis
    frames_df[frames_df.Angle == 0.00].apply(
        lambda row: op.element(
            "ElasticTimoshenkoBeam",
            row.UniqueName,
            row.PointI,
            row.PointJ,
            E,
            G,
            frame_props_df.loc[row.Prop, "Area"],
            frame_props_df.loc[row.Prop, "J"],
            frame_props_df.loc[row.Prop, "I33"],
            frame_props_df.loc[row.Prop, "I22"],
            frame_props_df.loc[row.Prop, "As3"],
            frame_props_df.loc[row.Prop, "As2"],
            determine_tag(row.Label),
            "-mass",
            M,
            massType,
        ),
        axis="columns",
    )

    # add all the frame elements which are oriented perpandicular to the major
    # axis by switching the properties in the two aixs

    frames_df[frames_df.Angle == 90.0].apply(
        lambda row: op.element(
            "ElasticTimoshenkoBeam",
            row.UniqueName,
            row.PointI,
            row.PointJ,
            E,
            G,
            frame_props_df.loc[row.Prop, "Area"],
            frame_props_df.loc[row.Prop, "J"],
            frame_props_df.loc[row.Prop, "I22"],
            frame_props_df.loc[row.Prop, "I33"],
            frame_props_df.loc[row.Prop, "As2"],
            frame_props_df.loc[row.Prop, "As3"],
            determine_tag(row.Label),
            "-mass",
            M,
            massType,
        ),
        axis="columns",
    )
    return


# ADD NODAL LOADS TO OPENSEES
def add_nodal_loads(loads_df):

    # create a plain load pattern
    op.pattern("Plain", 1, "Linear")

    # apply loads to the mdoel in opensees
    loads_df.apply(
        lambda row: op.load(
            *row[["UniqueName", "F1", "F2", "F3", "M1", "M2", "M3"]].tolist()
        ),
        axis="columns",
    )
    return


# PERFORM MODAL ANALYSIS IN OPENSEES
def modal_response(numEigen):

    # calculate eigenvalues
    eigenValues = op.eigen(
        "-genBandArpack", numEigen
    )  # can be either of: {'-genBandArpack', '-fullGenLapack'}
    return eigenValues


# PLOT MODE SHAPES OPTAINED FROM MODAL ANALYSIS IS OPENSEES
def plot_opensees_mode_shapes():
    opp.plot_model()
    for i in range(1, numEigen + 1):
        opp.plot_modeshape(i, 50)
    return


# ADD NON-LINEAR MOMENT HINGE TO THE MODEL
def add_beam_hinges(op, dict_of_hinges, dict_of_hinges_2):

    # obtain nonlinear hinge properties
    data = read_nonlinear_hinge_properties()

    #    dict_of_hinges = {real joint: (new joint, zero length element ID, orientation)}

    # Iterate through all the hinges and add the uniaxial material and zero length element to the opensees model
    for key, value in dict_of_hinges.items():
        node_R = key
        node_C = value[0]
        matTag = value[1]
        dirn = value[2]

        row = data.loc[data["Hinge NAME"] == dict_of_hinges_2[matTag]]
        K0 = row["K0"].values[0]
        as_Plus = row["as_Plus"].values[0]
        as_Neg = row["as_Neg"].values[0]
        My_Plus = row["My_Plus"].values[0]
        My_Neg = row["My_Neg"].values[0]
        Lamda_S = row["Lamda_S"].values[0]
        Lamda_C = row["Lamda_C"].values[0]
        Lamda_A = row["Lamda_A"].values[0]
        Lamda_K = row["Lamda_K"].values[0]
        c_S = row["c_S"].values[0]
        c_C = row["c_C"].values[0]
        c_A = row["c_A"].values[0]
        c_K = row["c_K"].values[0]
        theta_p_Plus = row["theta_p_Plus"].values[0]
        theta_p_Neg = row["theta_p_Neg"].values[0]
        theta_pc_Plus = row["theta_pc_Plus"].values[0]
        theta_pc_Neg = row["theta_pc_Neg"].values[0]
        Res_Pos = row["Res_Pos"].values[0]
        Res_Neg = row["Res_Neg"].values[0]
        theta_u_Plus = row["theta_u_Plus"].values[0]
        theta_u_Neg = row["theta_u_Neg"].values[0]
        D_Plus = row["D_Plus"].values[0]
        D_Neg = row["D_Neg"].values[0]
        nFactor = row["nFactor"].values[0]

        # define the uniaxial material
        op.uniaxialMaterial(
            "Bilin",
            matTag,
            K0,
            as_Plus,
            as_Neg,
            My_Plus,
            My_Neg,
            Lamda_S,
            Lamda_C,
            Lamda_A,
            Lamda_K,
            c_S,
            c_C,
            c_A,
            c_K,
            theta_p_Plus,
            theta_p_Neg,
            theta_pc_Plus,
            theta_pc_Neg,
            Res_Pos,
            Res_Neg,
            theta_u_Plus,
            theta_u_Neg,
            D_Plus,
            D_Neg,
            nFactor,
        )

        # Add the zero length element
        op.element(
            "zeroLength",
            matTag,
            node_R,
            node_C,
            "-mat",
            matTag,
            "-dir",
            dirn,
            "-doRayleigh",
            1,
        )

	# Constrain the nodes connecting the rero lengrth element - all DOFs
	# are constrained except the major bending

        op.equalDOF(node_R, node_C, 1, 2, 3, int(9 - dirn), 6)
        op.region(key, matTag)
    return


# SETUP TO RECORD ANALYSIS OUTPUT
def setup_recorders(
    dict_of_disp_nodes, dict_of_rxn_nodes, dict_of_hinges, initialOrTangent, parent_dir
):
    # set up node displacement recorders
    dict_of_disp_nodes = [61, 62, 63, 64, 65, 241, 242, 243, 244, 245]
    for node in dict_of_disp_nodes:
        op.recorder(
            "Node",
            "-file",
            f"node_{node}_disp_{initialOrTangent}.out",
            "-node",
            node,
            "-dof",
            1,
            2,
            3,
            4,
            5,
            6,
            "disp",
        )

    # set up node rxn recorders
    for node in dict_of_rxn_nodes.keys():
        op.recorder(
            "Node",
            "-file",
            f"node_{node}_rxn_{initialOrTangent}.out",
            "-node",
            node,
            "-dof",
            1,
            2,
            3,
            4,
            5,
            6,
            "reaction",
        )

    # setup rot spring recorders
    list_of_hinges = [
        20271,
        20275,
        20279,
        20283,
        20253,
        20672,
        20673,
        20674,
        20675,
        20676,
    ]
    for rec in list_of_hinges:
        op.recorder(
            "Element",
            "-file",
            f"ele_def_{rec}_{initialOrTangent}.out",
            "-ele",
            rec,
            "deformations",
        )
        op.recorder(
            "Element",
            "-file",
            f"ele_frc_{rec}_{initialOrTangent}.out",
            "-ele",
            rec,
            "-dof",
            1,
            2,
            3,
            4,
            5,
            6,
            "force",
        )
    return


# READ NONLINEAR PROPERTIES OF MOMENT HINGES FROM EXCEL SHEET AND FORMAT/ADD DATA FOR OPENSEES DEFINITION
def read_nonlinear_hinge_properties():
    """
    This function creates a uniaxial material spring with deterioration
    Spring follows: Bilinear Response based on Modified Ibarra Krawinkler Deterioration Model
    Written by: Dimitrios G. Lignos, Ph.D.

    matTag              integer tag identifying material
    K0                  elastic  stiffness
    as_Plus             strain hardening ratio for positive loading direction
    as_Neg              strain hardening ratio for negative loading direction
    My_Plus             effective yield strength for positive loading direction
    My_Neg              effective yield strength for negative loading direction (negative value)
    Lamda_S             Cyclic deterioration parameter for strength deterioration
    Lamda_C             Cyclic deterioration parameter for post-capping strength deterioration
    Lamda_A             Cyclic deterioration parameter for acceleration reloading stiffness deterioration (is not a deterioration mode for a component with Bilinear hysteretic response).
    Lamda_K             Cyclic deterioration parameter for unloading stiffness deterioration
    c_S                 rate of strength deterioration. The default value is 1.0.
    c_C                 rate of post-capping strength deterioration. The default value is 1.0.
    c_A                 rate of accelerated reloading deterioration. The default value is 1.0.
    c_K                 rate of unloading stiffness deterioration. The default value is 1.0.
    theta_p_Plus        pre-capping rotation for positive loading direction (often noted as plastic rotation capacity)
    theta_p_Neg         pre-capping rotation for negative loading direction (often noted as plastic rotation capacity) (positive value)
    theta_pc_Plus       post-capping rotation for positive loading direction
    theta_pc_Neg        post-capping rotation for negative loading direction (positive value)
    Res_Pos             residual strength ratio for positive loading direction
    Res_Neg             residual strength ratio for negative loading direction (positive value)
    theta_u_Plus        ultimate rotation capacity for positive loading direction
    theta_u_Neg         ultimate rotation capacity for negative loading direction (positive value)
    D_Plus              rate of cyclic deterioration in the positive loading direction (this parameter is used to create assymetric hysteretic behavior for the case of a composite beam). For symmetric hysteretic response use 1.0.
    D_Neg               rate of cyclic deterioration in the negative loading direction (this parameter is used to create assymetric hysteretic behavior for the case of a composite beam). For symmetric hysteretic response use 1.0.
    nFactor             elastic stiffness amplification factor, mainly for use with concentrated plastic hinge elements (optional, default = 0).
    """

    # Read properties from Excel
    data = pd.read_excel(
        os.path.join(
            os.path.dirname(os.getcwd()), "worksheets", "NL Properties Summary.xlsx"
        ),
        sheet_name="WUF hinge",
    )
    data.drop(
        columns=["IO (θy)", "LS (θy)", "CP (θy)", "Beam Standard Section?"],
        inplace=True,
    )

    # Generate properties for OpenSees Input
    data["K0"] = 1e7

    data["as_Plus"] = (
        (data["FU (k-in)"] - data["FY (k-in)"]) / data["DL (θy)"] / data["K0"]
    )
    data["as_Neg"] = data["as_Plus"]

    data["My_Plus"] = data["FY (k-in)"].astype(float)
    data["My_Neg"] = -data["FY (k-in)"].astype(float)

    data["Lamda_S"] = 1000.0  # ignore degradation
    data["Lamda_C"] = 1000.0  # ignore degradation
    data["Lamda_A"] = 1000.0  # ignore degradation
    data["Lamda_K"] = 1000.0  # ignore degradation

    data["c_S"] = 1.0
    data["c_C"] = 1.0
    data["c_A"] = 1.0
    data["c_K"] = 1.0

    data["theta_p_Plus"] = data["DL (θy)"] - data["My_Plus"] / data["K0"]
    data["theta_p_Neg"] = data["theta_p_Plus"]

    data["theta_pc_Plus"] = (data["DR (θy)"] - data["DL (θy)"]) / (1 - data["FR/FU"])
    data["theta_pc_Neg"] = data["theta_pc_Plus"]

    data["Res_Pos"] = data["FR/FU"]
    data["Res_Neg"] = data["FR/FU"]

    data["theta_u_Plus"] = 0.2
    data["theta_u_Neg"] = 0.2

    data["D_Plus"] = 1.0
    data["D_Neg"] = 1.0

    data["nFactor"] = 0.0

    return data


# RUN NLRHA USING RAYLEIGH DAMPING IN ETABS
def run_dynamic_analysis_w_rayleigh_damping(
    dict_of_hinges,
    dict_of_disp_nodes,
    dict_of_rxn_nodes,
    zeta,
    initialOrTangent="initial",
    parent_dir=os.getcwd(),
):

    # remove any existing analysis data
    op.wipeAnalysis()

    # define a time series to add the ground motion
    op.timeSeries(
        "Path", 2, "-dt", 0.01, "-filePath", "BM68elc.acc", "-factor", 3.0 * g
    )
    op.pattern("UniformExcitation", 2, 1, "-accel", 2)

    # uncomment/comment the code below to run bideirectional/unidirectional ground motion analysis
    #    op.timeSeries('Path', 3, '-dt', 0.01, '-filePath', 'BM68elc.acc', '-factor', 1.0*g)
    #    op.pattern('UniformExcitation', 3, 2, '-accel', 3)

    op.constraints("Transformation")
    op.numberer("RCM")
    op.system("UmfPack")
    #    op.test('NormDispIncr', 1e-2, 100000, 0, 0)
    op.test("EnergyIncr", 1e-4, 1e4, 0, 2)

    # available set of algorithms if the default fails
    backup_algos = {
        "Modified Newton w/ Initial Stiffness": ["ModifiedNewton", "-initial"],
        "Newton with Line Search": [
            "NewtonLineSearch",
            "tol",
            1e-3,
            "maxIter",
            1e5,
            "maxEta",
            10,
            "minEta",
            1e-2,
        ],
    }

    # define the default algorithm to be used for this analysis
    op.algorithm("KrylovNewton", "maxDim", 3)

    # define the integrator to be used for this analysis from the set of available integrators in opensees
    alpha = 0.67
    op.integrator(
        "HHT", alpha
    )  # can be either of: ('HHT', alpha), ('Newmark', 0.5, 0.25)

    # define the type of analysis to be performed
    op.analysis("Transient")

    # obtain the modal analysis
    eigenValues = modal_response(numEigen)
    # get periods from the modal analsis
    periods = 2 * math.pi / np.sqrt(eigenValues)

    w1 = eigenValues[0] ** 0.5
    w2 = (eigenValues[0] ** 0.5) / 0.384
    a0 = zeta * 2 * w1 * w2 / (w1 + w2)
    a1 = zeta * 2 / (w1 + w2)

    print("\nRayleigh Damping Coefficients:")
    print(f"\talpha: {a0}")
    print(f"\tbeta : {a1}\n")

    if initialOrTangent == "tangent":
        op.rayleigh(a0, a1, 0, 0)
    else:
        op.rayleigh(a0, 0, a1, 0)

    # setup to record analysis data
    setup_recorders(
        dict_of_disp_nodes,
        dict_of_rxn_nodes,
        dict_of_hinges,
        initialOrTangent,
        parent_dir,
    )

    total_run_time = 50  # seconds
    time_step = 0.01  # seconds
    total_num_of_steps = total_run_time / time_step

    # default initialization of constants to be used in the execution loop
    failed = 0
    time = 0
    algo = "Krylov-Newton"
    pbar = tqdm(total=total_num_of_steps)

    # execution loop
    # this execution loop tries to use krylov-Newton until it works, when this fails it iterate through set of
    # algorithms until it finds a solution, if it is not able to find a solution with any algorithm, the loop ends.
    # Whereas if it is able to find a solution with any of the algorithms, it switches back to krylov-Newton
    while time <= total_run_time and failed == 0:
        failed = op.analyze(1, time_step)

        if failed:
            print(f"\n{algo} failed. Trying other algorithms...")

            for alg, algo_args in backup_algos.items():
                print(f"\nTrying {alg}...")
                op.algorithm(*algo_args)
                failed = op.analyze(1, time_step)

                if failed:
                    continue
                else:
                    algo = "Krylov-Newton"
                    print(f"\n{alg} worked.\n\nMoving back to {algo}")
                    op.algorithm("KrylovNewton", "maxDim", 3)
                    break

            print("".center(100, "-"))

        pbar.update(1)
        time = op.getTime()

    op.wipe()

    # move output files to results directory
    list_of_out_files = [fname for fname in os.listdir() if fname.endswith(".out")]
    for fname in list_of_out_files:
        shutil.move(os.path.join(os.getcwd(), fname), os.path.join(parent_dir, fname))

    return periods, eigenValues


# COMPARES MODAL ANALYSIS PERIODS OBTAINED FROM ETABS AND OPENSEES
def perform_modal_analysis_and_comparison(etabs_periods):
    eigenValues = modal_response(len(etabs_periods))
    periods = 2 * math.pi / np.sqrt(eigenValues)

    # compare modal analysis results
    print("\nETABS Periods: ")
    print(etabs_periods)
    print("\nOpenSees Periods: ")
    print([round(n, 3) for n in list(periods)])
    print("\nAgreement: (ETABS Periods/OpenSees Periods)")
    print(etabs_periods / periods)
    op.wipeAnalysis()
    return


# EXECUTE ANALYSIS IN OPENSEES
def run_opensees_model(
    dict_of_hinges={},
    dict_of_disp_nodes={},
    dict_of_rxn_nodes={},
    zeta=0.05,
    initialOrTangent="",
    parent_dir=os.getcwd(),
):
    periods, eigenValues = run_dynamic_analysis_w_rayleigh_damping(
        dict_of_hinges,
        dict_of_disp_nodes,
        dict_of_rxn_nodes,
        zeta,
        initialOrTangent,
        parent_dir,
    )
    return periods, eigenValues





"""
    MIT License
    
    Copyright (c) 2020 OpenSeesPro
    
    Developed by:
        Ayush Singhania (ayushs@stanford.edu)
        Pearl Ranchal (ranchal@berkeley.edu)
      
    Publication:
        Goings, C. B., Singhania, A., Ranchal, P., Weaver B., 2020, “Industrial 
        Scale NLRH Analysis Using OpenSees and Comparison with Perform3D,” 
        Proceedings of 2020 SEAOC Virtual Convention, SEAOC, CA
        
    Description of the script - 
        This script is used to scrape data from ETABS and process it in a suitable 
        format to help generate the OpenSees model
"""

import pandas as pd
import math
from general_utilities import (
    get_database_table_for_all_load_cases_and_combos as get_dbtable,
    get_model_from_etabs,
)

# DEFINE DATAFRAME COLUMNS
JOINT_DATA_COLS = ["UniqueName", "X", "Y", "Z"]
FRAME_DATA_COLS = [
    "UniqueName",
    "Prop",
    "Story",
    "PointI",
    "PointJ",
    "PointIX",
    "PointIY",
    "PointIZ",
    "PointJX",
    "PointJY",
    "PointJZ",
    "Angle",
    "OffsetIX",
    "OffsetJX",
    "OffsetIY",
    "OffsetJY",
    "OffsetIZ",
    "OffsetJZ",
    "CardinalPt",
]
PT_LOADS_DATA_COLS = [
    "UniqueName",
    "LoadPattern",
    "Step",
    "CSys",
    "F1",
    "F2",
    "F3",
    "M1",
    "M2",
    "M3",
]
FRAME_PROP_COLS = [
    "Name",
    "Material",
    "Shape",
    "Area",
    "As2",
    "As3",
    "J",
    "I22",
    "I33",
    "S22Pos",
    "S33Pos",
    "Z22",
    "Z33",
    "R22",
    "R33",
    "I3Mod",
]
FRAME_PROP_COLS_2 = [
    "Area",
    "As2",
    "As3",
    "Torsion",
    "I22",
    "I33",
    "S22",
    "S33",
    "Z22",
    "Z33",
    "R22",
    "R33",
]


# EXTRACT JOINT CONNECTIVITY DATA FROM ETABS MODEL
def _get_joints(model):

    # extract the point object connectivity from etabs
    joints_df = get_dbtable("Point Object Connectivity", model)
    # extract the point object restraints from etabs
    joints_df["Restraints"] = joints_df.apply(
        lambda row: model.PointObj.GetRestraint(row.UniqueName)[0], axis="columns"
    )

    # get the maximum numberical joint ID, this is required to name various new objects while modeling nonlinearity in OpenSees
    max_jointID = max(
        joints_df[joints_df.UniqueName.str.isdigit()].UniqueName.astype("int").tolist()
    )
    constant_joint = 10 ** (int(math.log10(max_jointID)) + 2)
    constant_eleID = 2 * constant_joint

    dict_of_hinges = {}  # key - real joint; value - new joint, zero element ID
    dict_of_hinges_1 = (
        {}
    )  # key - 'N' + real joint; value - real joint, new joint, zero element ID
    list_new_joints = []

    # List of joints where nonlinearity is introduced (link defined in ETBAS
    # are connected to two joints one is digital say xx then other is Nxx). We
    # would first like to extract all joints with UniqueName non-Digit (prefix:
    # "N"). We would then use these nodes to create new node names for OpenSees
    # model using the constant_joint to name the ID and coordinate of the xx
    # joint to plate the new joint. (OpenSees need the two joints in the same
    # location for defining a zero-length element).

    dummy_joints = joints_df[~joints_df.UniqueName.str.isdigit()].UniqueName.tolist()

    # This loop goes through all the joints and modifies the dummy joint with
    # prefix "N" to the new joint in OpenSees

    for index, row in joints_df.iterrows():
        if row.UniqueName in dummy_joints:
            real_joint = int(str(row.UniqueName)[1:])
            new_joint = real_joint + constant_joint

            if (
                joints_df.loc[joints_df.UniqueName == str(real_joint), "X"]
                .astype(float)
                .values[0]
                == joints_df.loc[joints_df.UniqueName == row.UniqueName, "X"]
                .astype(float)
                .values[0]
            ):
                zle_dirn = 4
            else:
                zle_dirn = 5

            list_new_joints.append(new_joint)
            new_element = real_joint + constant_eleID
            joints_df.loc[index, "UniqueName"] = new_joint
            joints_df.loc[index, "X"] = joints_df[
                joints_df.UniqueName == str(real_joint)
            ]["X"].tolist()[0]
            joints_df.loc[index, "Y"] = joints_df[
                joints_df.UniqueName == str(real_joint)
            ]["Y"].tolist()[0]
            joints_df.loc[index, "Z"] = joints_df[
                joints_df.UniqueName == str(real_joint)
            ]["Z"].tolist()[0]
            dict_of_hinges[real_joint] = (new_joint, new_element, zle_dirn)
            dict_of_hinges_1["N" + str(real_joint)] = (
                real_joint,
                new_joint,
                new_element,
            )

    joints_df = joints_df.astype(
        {"UniqueName": int, "X": float, "Y": float, "Z": float}
    )
    return joints_df.copy(), dict_of_hinges, dict_of_hinges_1, list_new_joints


# EXTRACT NODAL LOADS FROM ETABS
def _get_pt_loads(model):

    joints_df, dict_of_hinges, dict_of_hinges_1, list_new_joints = _get_joints(model)

    # Initialize a dataframe to store the point loads
    pts_loads_df = pd.DataFrame(columns=PT_LOADS_DATA_COLS)

    # Iterate through the joints to get the loads from etabs if any
    for index, row in joints_df.iterrows():
        try:
            pt_loads_df = pd.DataFrame.from_dict(
                {
                    col: val
                    for (col, val) in zip(
                        PT_LOADS_DATA_COLS,
                        model.PointObj.GetLoadForce(str(row.UniqueName))[1:-1],
                    )
                }
            )
        except:
            continue

        pts_loads_df = pts_loads_df.append(pt_loads_df, ignore_index=True)

    pts_loads_df = pts_loads_df.astype(
        {
            "UniqueName": int,
        }
    )

    # filter out only the dead load
    pts_loads_df = pts_loads_df[pts_loads_df.LoadPattern == "Dead"].copy()

    return pts_loads_df.copy()


# EXTRACT FRAME CONNECTIVITY FROM ETABS
def _get_frames(dict_of_hinges_1, model):
    # Extract all frame labels from etabs
    frame_label_data = model.FrameObj.GetLabelNameList()
    frame_labels_df = pd.DataFrame.from_dict(
        {
            col: val
            for col, val in zip(
                ["UniqueName", "Label", "Story"], frame_label_data[1:-1]
            )
        }
    )

    # get all frame objects from etabs
    frame_data = model.FrameObj.GetAllFrames()
    frames_df = pd.DataFrame.from_dict(
        {col: val for (col, val) in zip(FRAME_DATA_COLS, frame_data[1:-1])}
    )
    frames_df["Label"] = frames_df["UniqueName"].replace(
        frame_labels_df.set_index("UniqueName")["Label"]
    )

    # Dictionary to get NL properties for the zero length element
    dict_of_hinges_2 = {}

    # update joint Names by iterating through each frames and chenging the names of both I & J nodes if required
    for index, row in frames_df.iterrows():
        # check if point I has been renmaed
        if row.PointI in dict_of_hinges_1.keys():
            dict_of_hinges_2[dict_of_hinges_1[row.PointI][2]] = frames_df.loc[
                index, "Prop"
            ]
            # replace the node with new joint
            frames_df.loc[index, "PointI"] = dict_of_hinges_1[row.PointI][1]

        # check if point J has been renamed
        if row.PointJ in dict_of_hinges_1.keys():
            dict_of_hinges_2[dict_of_hinges_1[row.PointJ][2]] = frames_df.loc[
                index, "Prop"
            ]
            # replace the node with new joint
            frames_df.loc[index, "PointJ"] = dict_of_hinges_1[row.PointJ][1]

    frames_df = frames_df.astype({"UniqueName": int, "PointI": int, "PointJ": int})
    return frames_df.copy(), dict_of_hinges_2


# EXTRACT NODAL MASSES FROM ETABS
def get_nodal_masses(model=None):

    # extract the assembled joint masses from etabs
    mass_df = get_dbtable("Assembled Joint Masses", model)
    mass_df = mass_df.astype(
        {
            "PointElm": int,
            "UX": float,
            "UY": float,
            "UZ": float,
            "RX": float,
            "RY": float,
            "RZ": float,
            "X": float,
            "Y": float,
            "Z": float,
        }
    )
    return mass_df.copy()


# EXTRACT FRAME SECTION PROPERTIES FROM ETABS
def get_frame_props(dict_of_hinges_1, model):

    frames_df, dict_of_hinges_2 = _get_frames(dict_of_hinges_1, model)

    # Initialize a dataframe to store the frame properties
    frame_props_df = pd.DataFrame(columns=FRAME_PROP_COLS_2)

    # iterate trough each fame to extract the frame properties if it does not already exists
    for index, row in frames_df.iterrows():

        # if the property exists then continue to next
        if row.Prop in frame_props_df.index:
            continue

        # if the property does not exist, extract from etabs and add to the dataframe using append
        frame_prop_df = pd.DataFrame.from_dict(
            {
                row.Prop: {
                    col: val
                    for (col, val) in zip(
                        FRAME_PROP_COLS_2, model.PropFrame.GetSectProps(row.Prop)[:-1]
                    )
                }
            },
            orient="index",
        )
        frame_props_df = frame_props_df.append(frame_prop_df)

    return frame_props_df.copy()


# EXTRACT FRAME SECTION PROPERTIES FROM ETBAS
def get_frame_props_from_db_table(model):

    FLOAT_COLS = [
        "Area",
        "As2",
        "As3",
        "J",
        "I22",
        "I33",
        "S22Pos",
        "S33Pos",
        "Z22",
        "Z33",
        "R22",
        "R33",
        "I3Mod",
    ]

    # extract the frame section properties from etabs
    frame_props_df = get_dbtable("Frame Section Property Definitions - Summary", model)[
        FRAME_PROP_COLS
    ]

    # update the properties as per modifiers defined in ETABS
    frame_props_df["I33"] = frame_props_df["I33"].astype(float) * frame_props_df[
        "I3Mod"
    ].astype(float)
    frame_props_df.set_index("Name", inplace=True)
    frame_props_df = frame_props_df.astype({col: "float" for col in FLOAT_COLS})
    return frame_props_df.copy()


# GET JOINTS WHERE REACTION NEEDS TO BE RECORDED AND WHERE DISPLACEMENT NEEDS TO BE RECORDED
def get_node_dicts(joints_df):

    # joints not on the lowest level are diaplcement nodes
    disp_joints_df = joints_df[joints_df.Z != joints_df.Z.min()].copy()

    # all joints at the lowest level are reaction nodes
    rxn_joints_df = joints_df[joints_df.Z == joints_df.Z.min()].copy()

    dict_of_disp_nodes = (
        disp_joints_df[["UniqueName", "X", "Y", "Z"]]
        .set_index("UniqueName")
        .T.to_dict()
    )
    dict_of_rxn_nodes = (
        rxn_joints_df[["UniqueName", "X", "Y", "Z"]].set_index("UniqueName").T.to_dict()
    )

    return dict_of_disp_nodes, dict_of_rxn_nodes


# EXTRACT MODAL ANALYSIS RESULTS FROM ETABS
def get_modal_results_from_etabs(model):

    # extract the Modal Participating Mass Ratio table from etabs
    df = get_dbtable("Modal Participating Mass Ratios", model)
    etabs_periods = [round(n, 3) for n in df.Period.astype(float).tolist()[:4]]
    return etabs_periods


# MAIN FUNCTION TO OBTAIN ALL REQUIRED DATA FROM ETABS TO CREATE A OPENSEES MODEL
def get_etabs_data(model=None, units=None):

    if model is None:
        model = get_model_from_etabs()
        model.SetPresentUnits(units)

    # call different methods written above to extract and process data from ETABS
    joints, hinges, hinges_1, new_joints = _get_joints(model)
    disp_nodes, rxn_nodes = get_node_dicts(joints.copy())
    pt_loads         = _get_pt_loads(model)
    frames, hinges_2 = _get_frames(hinges_1, model)
    mass             = get_nodal_masses(model)
    frame_props      = get_frame_props_from_db_table(model)
    etabs_periods    = get_modal_results_from_etabs(model)

    return (
        joints,
        pt_loads,
        frames,
        mass,
        frame_props,
        hinges,
        hinges_2,
        new_joints,
        disp_nodes,
        rxn_nodes,
        etabs_periods,
    )




from xcsi.csi import find_row, find_rows
from collections import defaultdict

def _apply_loads(csi, model, case: str=None):
    "LOAD CASE DEFINITIONS",
    "LOAD PATTERN DEFINITIONS",

    "JOINT LOADS - FORCE",
    "FRAME LOADS - DISTRIBUTED",
    "FRAME LOADS - GRAVITY",
    "FRAME LOADS - POINT",
    "CABLE LOADS - DISTRIBUTED",
    pass

class LoadPattern:
    def __init__(self, name, csi):
        self.name = name

        joint_sp = defaultdict(dict)
        for joint in find_rows(csi.get("JOINT LOADS - GROUND DISPLACEMENT", []),
                               LoadPat=name):
            for key in "U1", "U2", "U3", "R1", "R2", "R3":
                if key in joint:
                    joint_sp[joint["Joint"]][key] = joint[key]

        self._joint_sp = joint_sp


        joint_force = defaultdict(dict)
        for joint in find_rows(csi.get("JOINT LOADS - FORCE", []),
                               LoadPat=name):
            for key in "F1", "F2", "F3", "M1", "M2", "M3":
                if key in joint:
                    joint_force[joint["Joint"]][key] = joint[key]

        self._joint_force = joint_force

    def _add_frame_loads(self):
        name = self.name 
        csi = self._csi

        # Frame loads
        self._frame_loads = defaultdict(list)
        for assign in find_rows(csi.get("FRAME LOADS - DISTRIBUTED", []),
                               LoadPat=name):
            
            if assign.get("RelDistA", 0) != 0:
                raise ValueError("Relative distance not supported for distributed frame loads")
            if assign.get("RelDistB", 1) != 1:
                raise ValueError("Relative distance not supported for distributed frame loads")
            if assign.get("DistType", "RelDist") != "RelDist":
                raise ValueError("Only relative distance supported for distributed frame loads")
            
            try:
                direction = int(assign["Dir"]-1)
            except ValueError:
                raise ValueError(f"Unsupported direction {assign['Dir']} for frame load {assign['Frame']}")

            force = [0,0,0]
            moment = [0,0,0]
            if assign["Type"] == "Force":
                if assign["FOverLA"] != assign["FOverLB"]:
                    raise ValueError("FOverLA and FOverLB must be equal for distributed frame loads")
                
                force[direction] = assign["FOverLA"]
            elif assign["Type"] == "Moment":
                if assign["MOverLA"] != assign["MOverLB"]:
                    raise ValueError("MOverLA and MOverLB must be equal for distributed frame loads")

                raise NotImplementedError("Moment loads not implemented for frames")

            self._frame_loads[assign["Frame"]].append({
                "type": "Uniform",
                "basis": assign["CoordSys"].title(),
                "force": force,
                "moment": moment,
            })

        "FRAME LOADS - POINT"


    def apply(self, asm):
        """
        Apply the load pattern to the model.
        """
        model = asm.model
        dofs = ["U1", "U2", "U3", "R1", "R2", "R3"]

        ptag = asm.names.define("LoadPattern", "pattern", self.name)
        model.pattern("Plain", ptag, "Linear")
        for joint, sp in self._joint_sp.items():
            if not sp:
                continue
            for dof, value in sp.items():
                model.sp(
                    asm.names.identify("Joint", "node", joint),
                    dofs.index(dof)+1,
                    value,
                    pattern=ptag
                )

    def __str__(self):
        return f"LoadPattern(name={self.name})"


class LinearStaticAnalysis:
    def __init__(self, name, job):
        self.name = name
        self._csi = job._tree
        self._job = job


    def run(self, asm):
        """
        Run the linear static analysis case.
        """
        self._apply_loads(asm)

        # Run the analysis
        model = asm.model
        model.test("EnergyIncr", 1e-10, 10)
        model.system("Umfpack")
        model.constraints("Transformation")
        model.numberer("RCM")
        model.integrator("LoadControl", 1.0)
        model.algorithm("Newton")
        model.analysis("Static")
        model.analyze(1)
        
    def _apply_loads(self, asm):
        """
        Apply loads to the model for this linear static analysis case.
        """

        for assign in find_rows(self._csi.get("CASE - STATIC 1 - LOAD ASSIGNMENTS", []),
                       Case=self.name):
            pattern = self._job.pattern(assign["LoadName"])

            pattern.apply(asm)

    def __str__(self):
        return f"LinearStaticAnalysis(name={self.name})"
    
    def __repr__(self):
        return f"LinearStaticAnalysis(name={self.name})"

        
            
        


class NonlinearStaticAnalysis:
    def __init__(self, name, defs, csi, model):

        SolScheme = "Iterative Only"

        p04 = find_row(csi.get("CASE - STATIC 4 - NONLINEAR PARAMETERS", []),
                       Case=name)

        MaxIterNR = 10
        if p04 is not None:
            SolScheme = p04.get("SolScheme", SolScheme)

            if SolScheme != "Iterative Only":
                raise ValueError(f"Unsupported solution scheme: {SolScheme}")
            
            MaxIterNR = p04.get("MaxIterNR", MaxIterNR)
            ItConvTol = p04.get("ItConvTol", 1e-6)
            

        model.test()

    def apply_loads(self, model):
        """
        Apply loads to the model for this nonlinear static analysis case.
        """
        for assign in find_rows(self._csi.get("CASE - STATIC 2 - NONLINEAR LOAD APPLICATION", []),
                       Case=self.name):
            pattern = self._job.pattern(assign["LoadName"])



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#

from .utility import UnimplementedInstance, find_row

def create_points(csi, model, library, config, conv):
    log = []
    ndm = config["ndm"]
    ndf = config["ndf"]
#   dofs = config["dofs"]
    dofs = csi["ACTIVE DEGREES OF FREEDOM"][0]

    used = set()

    for node in csi["JOINT COORDINATES"]:

        if node["CoordSys"] != "GLOBAL":
            coordinates = tuple(node[i] if i in node else 0.0 for i in ("GlobalX", "GlobalY", "GlobalZ"))
            # log.append(UnimplementedInstance(f"Joint.Mass.CoordSys={node['CoordSys']}", node))
        else:
            coordinates = tuple(node[i] if i in node else 0.0 for i in ("XorR", "Y", "Z"))

        node_tag = conv.define("Joint", "node", node["Joint"])


        model.node(node_tag, coordinates)

        for i,v in enumerate(dofs.values()):
            if not v:
                model.fix(node_tag, dof=i+1)

    for node in csi.get("JOINT RESTRAINT ASSIGNMENTS", []):

        node_tag = conv.identify("Joint", "node", node["Joint"])
        # Only fix dofs that werent aready fixed globally
        # Note that dof keys look like UX, RY, etc, but in the restraint
        # table they look like U1, R2, etc
        model.fix(node_tag, tuple(int(node[f"{key[0]}{'XYZ'.find(key[1])+1}"]) if dofs[key] else 0 for key in dofs))


    for node in csi.get("JOINT ADDED MASS ASSIGNMENTS", []):
        mass = [node[f"Mass{i+1}"] for i in range(ndm)]
        mass = mass + [0.0]*(ndf-len(mass))
        node_tag = conv.identify("Joint", "node", node["Joint"])
        if node["CoordSys"] != "GLOBAL":
            conv.log(UnimplementedInstance(f"Joint.Mass.CoordSys={node['CoordSys']}", node))
        model.mass(node_tag, tuple(mass))

    
    for node in csi.get("JOINT ADDED MASS BY VOLUME ASSIGNMENTS", []):
        dens = find_row(csi.get("MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES",[]),
                        Material=node["Material"])["UnitMass"]
        vols = [node[f"Vol{i}"] for i in range(1,ndm+1) if f"Vol{i}" in node]
        vols = vols + [0.0]*(ndf-len(vols))
        mass = tuple(vol*dens for vol in vols)
        model.mass(node["Joint"], mass)

    used.add("JOINT COORDINATES")
    used.add("JOINT RESTRAINT ASSIGNMENTS")
    used.add("JOINT ADDED MASS ASSIGNMENTS")
    used.add("JOINT ADDED MASS BY VOLUME ASSIGNMENTS")


    log.extend( _apply_constraints(csi, model, conv) )

    used.add("JOINT CONSTRAINT ASSIGNMENTS")

    # TODO
    "JOINT MERGE NUMBER ASSIGNMENTS"

    return log


def _apply_constraints(sap, model, conv):
    log = []

    # The format of body dictionary is {'node number':'constraint name'}
    constraints = {}

    for constraint in  sap.get("JOINT CONSTRAINT ASSIGNMENTS", []):
        if "Type" in constraint and constraint["Type"] == "Body":
            # map node number to constraint
            constraints[constraint["Joint"]] = constraint["Constraint"]
        else:
            log.append(UnimplementedInstance("Joint.Constraint", constraint))

    # Sort the dictionary by body name and return a list [(node, body name)]
    constraints = list(sorted(constraints.items(), key=lambda x: x[1]))


    if len(constraints) > 0:
        nodes = []
        # Assign the first body name to the pointer
        pointer = constraints[0][1]

        # Traverse the tuple. If the second element in the tuple, the body
        # name, is the same as the pointer, then store the node number, 
        # into nodes.
        for node, constraint in constraints:
            if constraint == pointer:
                nodes.append(node)
            else:
                # First write the nodes in nodes to the body file
                for le in range(len(nodes)-1):
                    ni = conv.identify("Joint", "node", nodes[0])
                    nj = conv.identify("Joint", "node", nodes[le + 1])
                    model.eval(f"rigidLink beam {ni} {nj}\n")
                # Restore nodes and save the node that returns False.
                nodes = []
                nodes.append(node)
                # The pointer is changed to the new body name
                pointer = constraint

        # After the for loop ends, write the nodes in the nodes of the last loop to the body file.
        for le in range(len(nodes)-1):
            model.eval(f"rigidLink beam {nodes[0]} {nodes[le + 1]}\n")

    return log



import sees.reader.csi as csi


def make_schema(table, name):
    from genson import SchemaBuilder

    builder = SchemaBuilder()
    builder.add_schema({"type": "object", "properties": {}, "name": name})

    for row in table:
        builder.add_object(row)

    return builder.to_schema()



if __name__ == "__main__":
    import sys
    import json

    tables = {}
    for file in sys.argv[1:]:
        print(file, file=sys.stderr)
        with open(file, "r") as f:
            csi.load(f, append=tables)


#   print(json.dumps(tables))

    print(json.dumps({key: print(key, file=sys.stderr) or make_schema(row, key)
          for key, row in tables.items()}))





#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import math
from xcsi.csi import create_model, load, collect_outlines

def main():
    import sys

    with open(sys.argv[2], "r") as f:
        csi = load(f)


    if sys.argv[1][1] == "C":
        # Convert
        if "tcl" in sys.argv[1]:
            import xara 
            model = xara.Model(ndm=3, ndf=6, echo_file=sys.stdout)
            model = create_model(csi, model=model, verbose=False)
            sys.exit()

        else:
            model = create_model(csi, verbose=False)
            model.print("-json")
            sys.exit()


    model = create_model(csi, verbose=True)

    if sys.argv[1] == "-E":
        # Eigen
        import veux
        model.constraints("Transformation")
        if len(sys.argv) > 3:
            mode = int(sys.argv[3])
        else:
            mode = 1
        scale = 100
        e = model.eigen(mode)[-1]
        print(f"period = {2*math.pi/math.sqrt(e)}")
        veux.serve(veux.render_mode(model, mode, scale, vertical=3))


    elif sys.argv[1][:2] == "-V":

        # Visualize
        import veux
        outlines = collect_outlines(csi, model.frame_tags)
        artist = veux.create_artist(model, canvas="gltf", vertical=3,
                    model_config={
                        "frame_outlines": outlines
                    }
        )
        artist.draw_outlines()
        artist.draw_surfaces()


        if sys.argv[1] == "-Vo":
            artist.save(sys.argv[3])
        else:
            veux.serve(artist)

    elif sys.argv[1] == "-Vn":
        # Visualize
        from scipy.linalg import null_space
        model.constraints("Transformation")
        model.analysis("Static")
        K = model.getTangent().T
        v = null_space(K)[:,0] #, rcond=1e-8)
        print(v)


        u = {
            tag: [1000*v[dof-1] for dof in model.nodeDOFs(tag)]
            for tag in model.getNodeTags()
        }

        import veux
        veux.serve(veux.render(model, u, canvas="gltf", vertical=3))

    elif sys.argv[1] == "-Q":
        # Quiet conversion
        pass
    else:
        raise ValueError(f"Unknown operation {sys.argv[1]}")


if __name__ == "__main__":
    main()


"""
    MIT License
    
    Copyright (c) 2020 OpenSeesPro
    
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:
    
    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.
    
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    Developed by:
        Ayush Singhania (ayushs@stanford.edu)
        Pearl Ranchal (ranchal@berkeley.edu)
      
    Publication:
        Goings, C. B., Singhania, A., Ranchal, P., Weaver B., 2020, “Industrial 
        Scale NLRH Analysis Using OpenSees and Comparison with Perform3D,” 
        Proceedings of 2020 SEAOC Virtual Convention, SEAOC, CA
    
    Description of the script - 
        This is a supporting script for the main.py (user should execute main.py)
        This script is used to post-process the data generated through analyses
        in OpenSees.

"""

import os
import numpy as np
import pandas as pd

COLS = ["FX", "FY", "FZ", "MX", "MY", "MZ"]
COL_DICT = {i: col for i, col in zip(range(6), COLS)}


def post_process(initialOrTangent, dir_):
    fpath = os.path.join(dir_, f"ele_frc_20279_{initialOrTangent}.out")
    fpath2 = os.path.join(dir_, f"ele_def_20279_{initialOrTangent}.out")

    df = pd.DataFrame([s.split() for s in open(fpath, "r").readlines()])
    df["RY"] = pd.DataFrame([s.split() for s in open(fpath2, "r").readlines()])[0]
    df = df.astype(float).rename(columns=COL_DICT)

    ax = df.plot(x="RY", y="MY", grid=True, figsize=(15, 5))
    ax.set_axisbelow(True)

    df.to_excel(os.path.join(dir_, f"hinge_hyst-{initialOrTangent}.xlsx"))
    return df.copy()


def base_shear(dir_, dict_of_rxn_nodes, initialOrTangent):
    df_shear_x = pd.DataFrame(columns=[])
    df_shear_x["t"] = np.arange(0, 50.01, 0.01)
    df_shear_y = df_shear_x.copy()

    for rxn_node in dict_of_rxn_nodes:
        fpath = os.path.join(
            dir_, "node_" + str(rxn_node) + "_rxn_" + initialOrTangent + ".out"
        )
        df = pd.DataFrame([s.split() for s in open(fpath, "r").readlines()])
        df = df.astype(float).rename(columns=COL_DICT)

        df_shear_x[f"X - {rxn_node}"] = df.FX
        df_shear_y[f"Y - {rxn_node}"] = df.FY

    df_shear_x["Vx"] = df_shear_x.sum(axis=1)
    df_shear_y["Vy"] = df_shear_y.sum(axis=1)

    df_shear_x.to_excel(os.path.join(dir_, f"base shear x-{initialOrTangent}.xlsx"))
    df_shear_y.to_excel(os.path.join(dir_, f"base shear y-{initialOrTangent}.xlsx"))

    return




class Tables:
    def __init__(self, data: dict):
        self.data = data

    def any(self, table, **kwds)->dict:
        pass

    def get(self, table, _default, **kwds)->dict:
        for row in self.data:
            match = True
            for k, v in kwds.items():
                if k not in row or row[k] != v:
                    match = False
                    break

            if match:
                return row

    def all(self, table, **kwds)->list:
        rows = []
        for row in table:
            match = True
            for k, v in kwds.items():
                if k not in row or row[k] != v:
                    match = False
                    break

            if match:
                rows.append(row)

        return rows


_Tables = {
    "General": {
        "ACTIVE DEGREES OF FREEDOM",
        "ANALYSIS OPTIONS",
        "PROGRAM CONTROL",
        "PROJECT INFORMATION",
        "REBAR SIZES",
    },

    "Options": {
        "OPTIONS - COLORS - DISPLAY",
        "OPTIONS - COLORS - OUTPUT",
        "TABLES AUTOMATICALLY SAVED AFTER ANALYSIS",
    },

    "Connectivity": {
        "CONNECTIVITY - AREA",
        "CONNECTIVITY - CABLE",
        "CONNECTIVITY - FRAME",
        "CONNECTIVITY - LINK",
        "CONNECTIVITY - SOLID",
        "CONNECTIVITY - TENDON",
    },

    "Loads": {
        "LOAD CASE DEFINITIONS",
        "LOAD PATTERN DEFINITIONS",
        "MASS SOURCE",
        "LINK LOADS - GRAVITY",
        "JOINT LOADS - FORCE",
        "JOINT LOADS - GROUND DISPLACEMENT",

        "FRAME LOADS - DISTRIBUTED",
        "FRAME LOADS - GRAVITY",
        "FRAME LOADS - POINT",
        "FRAME LOADS - TEMPERATURE",
        "CABLE LOADS - DISTRIBUTED",
        "CABLE LOADS - TEMPERATURE",

        "SOLID LOADS - GRAVITY",
        "SOLID LOADS - PORE PRESSURE",
        "SOLID LOADS - SURFACE PRESSURE",
        "SOLID LOADS - TEMPERATURE",

        "AREA LOADS - GRAVITY",
        "AREA LOADS - PORE PRESSURE",
        "AREA LOADS - ROTATE",
        "AREA LOADS - SURFACE PRESSURE",
        "AREA LOADS - TEMPERATURE",
        "AREA LOADS - UNIFORM",

        "TENDON LOADS - TENSION FORCE OR STRESS",
    },

    "Material": {
        "MATERIAL PROPERTIES 01 - GENERAL",
        "MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES",

        "MATERIAL PROPERTIES 03A - STEEL DATA",
        "MATERIAL PROPERTIES 03B - CONCRETE DATA",
        "MATERIAL PROPERTIES 03C - ALUMINUM DATA",
        "MATERIAL PROPERTIES 03D - COLD FORMED DATA",
        "MATERIAL PROPERTIES 03E - REBAR DATA",
        "MATERIAL PROPERTIES 03F - TENDON DATA",
        "MATERIAL PROPERTIES 03G - OTHER DATA",
        "MATERIAL PROPERTIES 03J - COUPLED NONLINEAR VON MISES DATA",

        "MATERIAL PROPERTIES 04 - USER STRESS-STRAIN CURVES",
        "MATERIAL PROPERTIES 06 - DAMPING PARAMETERS",
        "MATERIAL PROPERTIES 09 - ACCEPTANCE CRITERIA",
    },

    "Joint": {
        "JOINT ADDED MASS ASSIGNMENTS",
        "JOINT ADDED MASS BY VOLUME ASSIGNMENTS",
        "JOINT BRIDGE OBJECT FLAGS",
        "JOINT CONSTRAINT ASSIGNMENTS",
        "JOINT COORDINATES",
        "JOINT LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "JOINT MERGE NUMBER ASSIGNMENTS",
        "JOINT PATTERN ASSIGNMENTS",
        "JOINT PATTERN DEFINITIONS",
        "JOINT RESTRAINT ASSIGNMENTS",
        "JOINT SPRING ASSIGNMENTS 1 - UNCOUPLED",
        "JOINT VEHICLE RESPONSE COMPONENT OVERWRITES",
    },

    "Point": {
        "LINK BRIDGE OBJECT FLAGS",
        "LINK FREQUENCY DEPENDENT PROPERTIES 01 - GENERAL",
        "LINK FREQUENCY DEPENDENT PROPERTIES 02 - DETAILS",
        "LINK LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "LINK LOCAL AXES ASSIGNMENTS 2 - ADVANCED",
        "LINK PROPERTY ASSIGNMENTS",
        "LINK PROPERTY DEFINITIONS - BRIDGE OBJECT FLAGS",
        "LINK PROPERTY DEFINITIONS 01 - GENERAL",
        "LINK PROPERTY DEFINITIONS 02 - LINEAR",
        "LINK PROPERTY DEFINITIONS 03 - MULTILINEAR",
        "LINK PROPERTY DEFINITIONS 04 - DAMPER",
        "LINK PROPERTY DEFINITIONS 05 - GAP",
        "LINK PROPERTY DEFINITIONS 06 - HOOK",
        "LINK PROPERTY DEFINITIONS 07 - RUBBER ISOLATOR",
        "LINK PROPERTY DEFINITIONS 08 - SLIDING ISOLATOR",
        "LINK PROPERTY DEFINITIONS 10 - PLASTIC (WEN)",
        "LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC",
        "LINK VEHICLE RESPONSE COMPONENT OVERWRITES",
    },

    "Frame": {
        "FRAME ADDED MASS ASSIGNMENTS",
        "FRAME AUTO MESH ASSIGNMENTS",
        "FRAME BRIDGE OBJECT FLAGS",
        "FRAME DESIGN PROCEDURES",
        "FRAME INSERTION POINT ASSIGNMENTS",
        "FRAME JOINT OFFSETS - BRIDGE OBJECT FLAGS",
        "FRAME LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        
        "FRAME P-DELTA FORCE ASSIGNMENTS",
        "FRAME PROPERTY MODIFIERS",
        "FRAME SPRING ASSIGNMENTS",

        # Hinges and releases
        "FRAME HINGE ASSIGNS 00 - HINGE DISTRIBUTION TYPE",
        "FRAME HINGE ASSIGNS 02 - USER DEFINED PROPERTIES",
        "FRAME HINGE ASSIGNS 09 - HINGE OVERWRITES",
        "FRAME RELEASE ASSIGNMENTS 1 - GENERAL",
        "FRAME RELEASE ASSIGNMENTS 2 - PARTIAL FIXITY",

        "FRAME LOAD TRANSFER OPTIONS",
        
        "FRAME OFFSET ALONG LENGTH ASSIGNMENTS",
        "FRAME OUTPUT STATION ASSIGNMENTS",
        "FRAME TENSION AND COMPRESSION LIMITS",
        "FRAME VEHICLE RESPONSE COMPONENT OVERWRITES",
        "FRAME SECTION ASSIGNMENTS",

        "FRAME SECTION PROPERTIES - BRIDGE OBJECT FLAGS",
        "FRAME SECTION PROPERTIES 01 - GENERAL",
        "FRAME SECTION PROPERTIES 02 - CONCRETE COLUMN",
        "FRAME SECTION PROPERTIES 05 - NONPRISMATIC",
        "FRAME SECTION PROPERTIES 06 - POLYGON DATA",
        "FRAME SECTION PROPERTIES 13 - TIME DEPENDENT",
    },

    "Shell": {
        "AREA LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "AREA LOCAL AXES ASSIGNMENTS 2 - ADVANCED",
        "AREA ADDED MASS ASSIGNMENTS",

        "AREA AUTO MESH ASSIGNMENTS",
        "AREA BRIDGE OBJECT FLAGS",
        "AREA EDGE CONSTRAINT ASSIGNMENTS",

        "AREA OVERWRITES - JOINT OFFSETS",
        "AREA OVERWRITES - THICKNESS",

        "AREA SECTION ASSIGNMENTS",
        "AREA SECTION PROPERTIES - BRIDGE OBJECT FLAGS",
        "AREA SECTION PROPERTIES",
        "AREA SECTION PROPERTY - TIME DEPENDENT",
        "AREA SECTION PROPERTY DESIGN PARAMETERS",
        "AREA SPRING ASSIGNMENTS",
        "AREA STIFFNESS MODIFIERS",
    },

    "Solid": {
        "SOLID AUTO MESH ASSIGNMENTS",
        "SOLID LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "SOLID PROPERTY ASSIGNMENTS",
        "SOLID PROPERTY DEFINITIONS",
        "SOLID SPRING ASSIGNMENTS",
    },

    "Cable": {
        "CABLE OUTPUT STATION ASSIGNMENTS",
        "CABLE SECTION ASSIGNMENTS",
        "CABLE SECTION DEFINITIONS",
        "CABLE SHAPE DATA",
    },
    "Bridge": {
        "BRIDGE ABUTMENT DEFINITIONS",
        "BRIDGE BEARING DEFINITIONS",
        "BRIDGE BENT DEFINITIONS 1 - GENERAL",
        "BRIDGE BENT DEFINITIONS 2 - COLUMN DATA",
        "BRIDGE DESIGN PREFERENCES - AASHTO LRFD 2020",
        "BRIDGE DESREQSUPER 01 - GENERAL",
        "BRIDGE DESREQSUPER 02 - STATION RANGES",
        "BRIDGE DESREQSUPER 03 - DEMAND SETS",
        "BRIDGE DESREQSUPER 04 - LLDF - AASHTOLRFD20",
        "BRIDGE DESREQSUPER 05 - PARAM - ALL CODES - CBOXTENDON",
        "BRIDGE DESREQSUPER 09 - PARAM - AASHTOLRFD20 - CBOX2STRESS",
        "BRIDGE DESREQSUPER 10 - PARAM - AASHTOLRFD20 - CBOX2SHEAR",
        "BRIDGE DESREQSUPER 11 - PARAM - AASHTOLRFD20 - CBOX2FLEXURE",
        "BRIDGE DIAPHRAGM DEFINITIONS",
        "BRIDGE ERECTION DEFINITION 01 - TRAVELER",
        "BRIDGE GROUP DEFINITION 01 - GENERAL",
        "BRIDGE GROUP DEFINITION 02 - SECTION",
        "BRIDGE GROUP DEFINITION 08 - SUPPORT STRUCTURE",
        "BRIDGE GROUP DEFINITION 09 - BEARING - SUPPORT",
        "BRIDGE GROUP DEFINITION 14 - MIXED",
        "BRIDGE LAYOUT LINE 1 - GENERAL",
        "BRIDGE LAYOUT LINE 2 - HORIZONTAL LAYOUT DATA",
        "BRIDGE LAYOUT LINE 3 - VERTICAL LAYOUT DATA",
        "BRIDGE LOAD DEFINITIONS 02 - LINE",
        "BRIDGE LOAD DEFINITIONS 03 - AREA",
        "BRIDGE OBJECT DEFINITIONS 01 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 02 - REFERENCE LINE",
        "BRIDGE OBJECT DEFINITIONS 03 - SPANS 1 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 06 - ABUTMENTS",
        "BRIDGE OBJECT DEFINITIONS 07 - BENTS",
        "BRIDGE OBJECT DEFINITIONS 09 - SUPER ELEVATION 1 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 11 - PRESTRESS 1 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 12 - PRESTRESS 2 - VERTICAL LAYOUT",
        "BRIDGE OBJECT DEFINITIONS 13 - PRESTRESS 3 - HORIZONTAL LAYOUT",
        "BRIDGE OBJECT DEFINITIONS 14 - DIAPHRAGMS",
        "BRIDGE OBJECT DEFINITIONS 15 - UPDATE DATA",
        "BRIDGE OBJECT DEFINITIONS 21 - TEMPERATURE LOADS",
        "BRIDGE OBJECT DEFINITIONS 26 - LINE LOADS",
        "BRIDGE OBJECT DEFINITIONS 27 - AREA LOADS",
        "BRIDGE OBJECT DEFINITIONS 28 - SPAN SEGMENTS",
        "BRIDGE OBJECT DEFINITIONS 31 - WIND LOADS",
        "BRIDGE OBJECT DEFINITIONS 32 - WIND LOAD - LOADPATTERN LIST",
        "BRIDGE PREFERENCES",
        "BRIDGE RATEREQSUPER 01 - GENERAL",
        "BRIDGE RATEREQSUPER 02 - STATION RANGES",
        "BRIDGE RATEREQSUPER 03 - DEMAND SETS",
        "BRIDGE RATEREQSUPER 07 - PARAM - AASHTORATE18 - CBOX2SHEAR",
        "BRIDGE RATEREQSUPER 08 - PARAM - AASHTORATE18 - CBOX2FLEXURE",
        "BRIDGE RATEREQSUPER 09 - PARAM - AASHTORATE18 - CBOX2MINREBAR",
        "BRIDGE RATEREQSUPER 20 - PARAM - AASHTORATE18 - CBOX2SERV",
        "BRIDGE RATING PREFERENCES - AASHTO RATING 2018",
        "BRIDGE RESPONSE",
        "BRIDGE SECTION CUTS 01 - GENERAL",
        "BRIDGE SECTION CUTS 02 - GROUPS",
        "BRIDGE SECTION CUTS 03 - STRESS POINTS",
        "BRIDGE SECTION CUTS 04 - GIRDER DATA - GENERAL",
        "BRIDGE SECTION CUTS 05 - GIRDER DATA - GROUPS",
        "BRIDGE SECTION DEFINITIONS - GIRDER OUTPUT LOCATION DATA",
        "BRIDGE SECTION DEFINITIONS 02 - CONCRETE BOX GIRDER",
        "BRIDGE SECTION LOAD DEFINITIONS 05 - BARRIER",
        "BRIDGE SECTION LOAD DEFINITIONS 07 - WEARING SURFACE",
        "BRIDGE SECTION LOAD DEFINITIONS 08 - TEMPERATURE CHANGE",
        "BRIDGE SEISMIC DESIGN PREFERENCES - AASHTO SEISMIC 2011",
    },
    "Other": {
        "AUTO WAVE 3 - WAVE CHARACTERISTICS - GENERAL",
        "AUTO WIND - AASHTO 2020",
        "AUTO WIND - LIVE LOAD",
        "AUTO WIND ANGLES",
        "AUTO WIND EXPOSURE FOR HORIZONTAL DIAPHRAGMS",


        "CASE - BUCKLING 1 - GENERAL",
        "CASE - BUCKLING 2 - LOAD ASSIGNMENTS",
        "CASE - DIRECT HISTORY 1 - GENERAL",
        "CASE - DIRECT HISTORY 2 - LOAD ASSIGNMENTS",
        "CASE - DIRECT HISTORY 3 - VISCOUS PROPORTIONAL DAMPING",
        "CASE - DIRECT HISTORY 4 - INTEGRATION PARAMETERS",
        "CASE - DIRECT HISTORY 5 - NONLINEAR PARAMETERS",
        "CASE - DIRECT HISTORY 6 - MODAL PROPORTIONAL DAMPING",
        "CASE - MODAL 1 - GENERAL",
        "CASE - MODAL 3 - LOAD ASSIGNMENTS - RITZ",
        "CASE - MODAL HISTORY 1 - GENERAL",
        "CASE - MODAL HISTORY 2 - LOAD ASSIGNMENTS",
        "CASE - MODAL HISTORY 4 - PROPORTIONAL DAMPING",
        "CASE - MODAL HISTORY 5 - DAMPING OVERRIDES",
        "CASE - MODAL HISTORY 6 - NONLINEAR PARAMETERS",
        "CASE - MOVING LOAD 1 - LANE ASSIGNMENTS",
        "CASE - MOVING LOAD 2 - LANES LOADED",
        "CASE - MOVING LOAD 3 - MULTILANE FACTORS",
        "CASE - MULTISTEP STATIC 1 - LOAD ASSIGNMENTS",
        "CASE - RESPONSE SPECTRUM 1 - GENERAL",
        "CASE - RESPONSE SPECTRUM 2 - LOAD ASSIGNMENTS",
        
        "CASE - STATIC 1 - LOAD ASSIGNMENTS",
        "CASE - STATIC 2 - NONLINEAR LOAD APPLICATION",
        "CASE - STATIC 4 - NONLINEAR PARAMETERS",
        "CASE - STATIC 5 - NONLINEAR STAGE DEFINITIONS",
        "CASE - STATIC 6 - NONLINEAR STAGE DATA",

        "CASE - STEADY STATE 1 - GENERAL",
        "CASE - STEADY STATE 2 - LOAD ASSIGNMENTS",
        "CASE - STEADY STATE 3 - ADDED FREQUENCIES GENERAL",
        "CASE - STEADY STATE 5 - ADDED SPECIFIED FREQUENCIES",
        "CASE - STEADY STATE 6 - CONSTANT DAMPING",
        "COMBINATION DEFINITIONS",

        "CONSTRAINT DEFINITIONS - BODY",
        "CONSTRAINT DEFINITIONS - BRIDGE OBJECT FLAGS",
        "CONSTRAINT DEFINITIONS - DIAPHRAGM",
        "CONSTRAINT DEFINITIONS - EQUAL",
        "CONSTRAINT DEFINITIONS - ROD",
        "COORDINATE SYSTEMS",
        "DATABASE DOCUMENTATION",
        "DATABASE FORMAT TYPES",
        "FLOATING LANE DEFINITION DATA",

        "FUNCTION - PLOT FUNCTIONS",
        "FUNCTION - POWER SPECTRAL DENSITY - USER",
        "FUNCTION - RESPONSE SPECTRUM - AASHTO 2007",
        "FUNCTION - RESPONSE SPECTRUM - USER",
        "FUNCTION - STEADY STATE - USER",
        "FUNCTION - TIME HISTORY - FROM FILE",
        "FUNCTION - TIME HISTORY - RAMP",
        "FUNCTION - TIME HISTORY - SINE",
        "FUNCTION - TIME HISTORY - USER",

        "GENERALIZED DISPLACEMENT DEFINITIONS 1 - TRANSLATIONAL",
        "GENERALIZED DISPLACEMENT DEFINITIONS 2 - ROTATIONAL",
        "GRID LINES",

        "GROUPS 1 - DEFINITIONS",
        "GROUPS 2 - ASSIGNMENTS",
        "GROUPS 4 - BRIDGE OBJECT FLAGS",

        "HINGES DEF 02 - NONINTERACTING - DEFORM CONTROL - GENERAL",
        "HINGES DEF 03 - NONINTERACTING - DEFORM CONTROL - FORCE-DEFORM",
        "HINGES DEF 04 - NONINTERACTING - DEFORM CONTROL - ACCEPTANCE",

        "LANE DEFINITION DATA",

        "NAMED SETS - DATABASE TABLES 1 - GENERAL",
        "NAMED SETS - DATABASE TABLES 2 - SELECTIONS",
        "OVERWRITES - CONCRETE DESIGN - ACI 318-08/IBC2009",
        "OVERWRITES - STEEL DESIGN - AISC360-05-IBC2006",

        "PREFERENCES - ALUMINUM DESIGN - AA 2015",
        "PREFERENCES - COLD FORMED DESIGN - AISI-16",
        "PREFERENCES - COLD FORMED DESIGN - AISI-ASD96",
        "PREFERENCES - CONCRETE DESIGN - AASHTO LRFD 2014",
        "PREFERENCES - CONCRETE DESIGN - ACI 318-08/IBC2009",
        "PREFERENCES - CONCRETE SHELL DESIGN - EUROCODE 2-2004",
        "PREFERENCES - DIMENSIONAL",
        "PREFERENCES - STEEL DESIGN - AASHTO-LRFD-2020",
        "PREFERENCES - STEEL DESIGN - AISC360-05-IBC2006",

        "SECTION CUTS 1 - GENERAL",

        "TENDON BRIDGE OBJECT FLAGS",
        "TENDON LAYOUT DATA 01 - GENERAL",
        "TENDON LAYOUT DATA 02 - SEGMENTS",
        "TENDON SECTION ASSIGNMENTS",
        "TENDON SECTION DEFINITIONS - BRIDGE OBJECT FLAGS",
        "TENDON SECTION DEFINITIONS",

        "VEHICLES 2 - GENERAL VEHICLES 1 - GENERAL",
        "VEHICLES 3 - GENERAL VEHICLES 2 - LOADS",
        "VEHICLES 4 - VEHICLE CLASSES",
    }
}



"""
    MIT License
    
    Copyright (c) 2020 OpenSeesPro

    Developed by:
        Ayush Singhania (ayushs@stanford.edu)
        Pearl Ranchal (ranchal@berkeley.edu)

    Description of this script - 
        This is the main script that the user is supposed to execute to use E20
        library. All other files are supporting scripts. This script first looks
        for a result folder and if not present creats it. Then it extracts data
        from ETBAS using the etabs_utilities function. Followed by execution of 
        setup_opensees_model method to generate the opensees model.Finally it 
        executes the analysis in OpenSees using the run_opensees_model method.

"""

from .general_utilities import start_time, end_time
from .etabs_utilities import get_etabs_data
from .opensees_utilities import (
    setup_opensees_model,
    perform_modal_analysis_and_comparison,
    run_opensees_model,
)
from .opensees_postprocessor import post_process, base_shear
import time
import os

if __name__ == "__main__":
    start = start_time()
    working_dir = os.path.join(os.path.dirname(os.getcwd()), "results")

    if not os.path.exists(working_dir):
        os.makedirs(working_dir)

    print("".center(100, "-"))
    print(":: GET ETABS MODEL DATA ::".center(100))
    print("".center(100, "-"))
    (
        joints_df,
        pts_loads_df,
        frames_df,
        mass_df,
        frame_props_df,
        dict_of_hinges,
        dict_of_hinges_2,
        list_new_joints,

        dict_of_disp_nodes,
        dict_of_rxn_nodes,
        etabs_periods,
    ) = get_etabs_data(units=3)
    print("Done!\n")

    print("".center(100, "-"))
    print(":: SET UP OPENSEES MODEL USING ETABS DATA ::".center(100))
    print("".center(100, "-"))
    setup_opensees_model(
        joints_df,
        frames_df,
        frame_props_df,
        pts_loads_df,
        mass_df,
        dict_of_hinges,
        dict_of_hinges_2,
        list_new_joints,
    )
    print("OpenSees Model Created!")
    end_time(start, final=False)

    print("".center(100, "-"))
    print(":: RUN OPENSEES MODEL ::".center(100))
    print("".center(100, "-"))

    print(f"Destination Directory: {working_dir}")
    time.sleep(1)

    # PERFORM MODAL ANALYSIS COMPARISON B/W ETABS AND OPENSEES
    perform_modal_analysis_and_comparison(etabs_periods)

    # RUN OPENSEES MODEL
    zeta = 0.05
    initialOrTangent = "tangent"
    periods, eigenValues = run_opensees_model(
        dict_of_hinges,
        dict_of_disp_nodes,
        dict_of_rxn_nodes,
        zeta,
        initialOrTangent,
        working_dir,
    )

    # POST-PROCESS ANALYSIS DATA
    df = post_process(initialOrTangent, working_dir)
    base_shear(working_dir, dict_of_rxn_nodes, initialOrTangent)

    # FINISH
    end_time(start)
    time.sleep(1.0)

